import pandas as pd
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
    _update_dataframe_datatypes,
    _create_dataframe,
    resolve_workspace_id,
    resolve_capacity_id,
)
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID
from typing import List, Optional


@log
def resolve_deployment_pipeline_id(deployment_pipeline: str | UUID) -> UUID:
    """
    Obtains the Id for a given deployment pipeline.

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    uuid.UUID
        The deployment pipeline Id.
    """

    if _is_valid_uuid(deployment_pipeline):
        return deployment_pipeline
    else:
        dfP = list_deployment_pipelines()
        dfP_filt = dfP[dfP["Deployment Pipeline Name"] == deployment_pipeline]
        if len(dfP_filt) == 0:
            raise ValueError(
                f"{icons.red_dot} The '{deployment_pipeline}' deployment pipeline is not valid."
            )
        return dfP_filt["Deployment Pipeline Id"].iloc[0]


@log
def resolve_stage_id(deployment_pipeline_id: UUID, stage: str | UUID):

    dfPS = list_deployment_pipeline_stages(deployment_pipeline=deployment_pipeline_id)

    if _is_valid_uuid(stage):
        dfPS_filt = dfPS[dfPS["Deployment Pipeline Stage Id"] == stage]
    else:
        dfPS_filt = dfPS[dfPS["Deployment Pipeline Stage Name"] == stage]
    if dfPS.empty:
        raise ValueError(
            f"{icons.red_dot} The '{stage}' stage does not exist within the '{deployment_pipeline_id}' deployment pipeline."
        )
    return dfPS_filt["Deployment Pipeline Stage Id"].iloc[0]


@log
def list_deployment_pipelines() -> pd.DataFrame:
    """
    Shows a list of deployment pipelines the user can access.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipelines <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipelines>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of deployment pipelines the user can access.
    """

    columns = {
        "Deployment Pipeline Id": "string",
        "Deployment Pipeline Name": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/deploymentPipelines",
        status_codes=200,
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Deployment Pipeline Id": v.get("id"),
                    "Deployment Pipeline Name": v.get("displayName"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_deployment_pipeline_stages(deployment_pipeline: str | UUID) -> pd.DataFrame:
    """
    Shows the specified deployment pipeline stages.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Stages <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-stages>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the specified deployment pipeline stages.
    """

    columns = {
        "Deployment Pipeline Stage Id": "string",
        "Deployment Pipeline Stage Name": "string",
        "Order": "int",
        "Description": "string",
        "Workspace Id": "string",
        "Workspace Name": "string",
        "Public": "bool",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages",
        status_codes=200,
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Deployment Pipeline Stage Id": v.get("id"),
                    "Deployment Pipeline Stage Name": v.get("displayName"),
                    "Description": v.get("description"),
                    "Order": v.get("order"),
                    "Workspace Id": v.get("workspaceId"),
                    "Workspace Name": v.get("workspaceName"),
                    "Public": v.get("isPublic"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_deployment_pipeline_stage_items(
    deployment_pipeline: str | UUID,
    stage: str | UUID,
) -> pd.DataFrame:
    """
    Shows the supported items from the workspace assigned to the specified stage of the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Stage Items <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-stage-items>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    stage : str | uuid.UUID
        The deployment pipeline stage name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the supported items from the workspace assigned to the specified stage of the specified deployment pipeline.
    """

    columns = {
        "Deployment Pipeline Stage Item Id": "string",
        "Deployment Pipeline Stage Item Name": "string",
        "Item Type": "string",
        "Source Item Id": "string",
        "Target Item Id": "string",
        "Last Deployment Time": "string",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    stage_id = resolve_stage_id(deployment_pipeline_id, stage)

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages/{stage_id}/items",
        status_codes=200,
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Deployment Pipeline Stage Item Id": v.get("itemId"),
                    "Deployment Pipeline Stage Item Name": v.get("itemDisplayName"),
                    "Item Type": v.get("itemType"),
                    "Source Item Id": v.get("sourceItemId"),
                    "Target Item Id": v.get("targetItemId"),
                    "Last Deployment Time": v.get("lastDeploymentTime"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_deployment_pipeline_role_assignments(
    deployment_pipeline: str | UUID,
) -> pd.DataFrame:
    """
    Shows the role assignments for the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the role assignments for the specified deployment pipeline.
    """

    columns = {
        "Role": "string",
        "Principal Id": "string",
        "Principal Type": "string",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/roleAssignments",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            principal = v.get("principal", {})
            rows.append(
                {
                    "Role": v.get("role"),
                    "Principal Id": principal.get("id"),
                    "Principal Type Name": principal.get("type"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_deployment_pipeline(
    deployment_pipeline: str | UUID,
):
    """
    Deletes the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - Delete Deployment Pipeline <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/delete-deployment-pipeline>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    """

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}",
        method="delete",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{deployment_pipeline}' deployment pipeline has been deleted successfully."
    )


@log
def list_deployment_pipeline_operations(
    deployment_pipeline: str | UUID,
) -> pd.DataFrame:
    """
    Shows the operations for the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Operations <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-operations>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the operations for the specified deployment pipeline.
    """

    columns = {
        "Operation Id": "string",
        "Type": "string",
        "Status": "string",
        "Last Updated Time": "string",
        "Execution Start Time": "datetime_coerce",
        "Execution End Time": "datetime_coerce",
        "Source Stage Id": "string",
        "Target Stage Id": "string",
        "Note": "string",
        "New Items Count": "int",
        "Different Items Count": "int",
        "No Difference Items Count": "int",
        "Performed By Id": "string",
        "Performed By Type": "string",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/operations",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            p = v.get("preDeploymentDiffInformation", {})
            rows.append(
                {
                    "Operation Id": v.get("id"),
                    "Type": v.get("type"),
                    "Status": v.get("status"),
                    "Last Updated Time": v.get("lastUpdatedTime"),
                    "Execution Start Time": v.get("executionStartTime"),
                    "Execution End Time": v.get("executionEndTime"),
                    "Source Stage Id": v.get("sourceStageId"),
                    "Target Stage Id": v.get("targetStageId"),
                    "Note": v.get("note", {}).get("content"),
                    "New Items Count": p.get("newItemsCount"),
                    "Different Items Count": p.get("differentItemsCount"),
                    "No Difference Items Count": p.get("noDifferenceItemsCount"),
                    "Performed By Id": v.get("performedBy", {}).get("id"),
                    "Performed By Type": v.get("performedBy", {}).get("type"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def unassign_workspace_from_stage(
    deployment_pipeline: str | UUID,
    stage: str | UUID,
):
    """
    Unassigns the workspace from the specified stage of the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - Unassign Workspace From Stage <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/unassign-workspace-from-stage>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    stage : str | uuid.UUID
        The deployment pipeline stage name or ID.
    """

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    stage_id = resolve_stage_id(deployment_pipeline_id, stage)

    _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages/{stage_id}/unassignWorkspace",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The workspace has been unassigned from the '{stage}' stage of the '{deployment_pipeline}' deployment pipeline successfully."
    )


@log
def assign_workspace_to_stage(
    deployment_pipeline: str | UUID,
    stage: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Unassigns the workspace from the specified stage of the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - Assign Workspace To Stage <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/assign-workspace-to-stage>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    stage : str | uuid.UUID
        The deployment pipeline stage name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    stage_id = resolve_stage_id(deployment_pipeline_id, stage)
    workspace_id = resolve_workspace_id(workspace=workspace)

    payload = {"workspaceId": workspace_id}

    _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages/{stage_id}/assignWorkspace",
        method="post",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The workspace has been assigned to the '{stage}' stage of the '{deployment_pipeline}' deployment pipeline successfully."
    )


@log
def deploy_stage_content(
    deployment_pipeline: str | UUID,
    source_stage_id: UUID,
    target_stage_id: UUID,
    items: Optional[dict | List[dict]] = None,
    note: Optional[str] = None,
    allow_cross_region_deployment: Optional[bool] = False,
    capacity: Optional[str | UUID] = None,
    workspace_name: Optional[str] = None,
):
    """
    Deploys items from the specified stage of the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - Deploy Stage Content <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/deploy-stage-content>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    source_stage_id : uuid.UUID
        The source deployment pipeline stage ID.
    target_stage_id : uuid.UUID
        The target deployment pipeline stage ID.
    items : dict | List[dict], default=None
        A list of items to deploy. Each item should be a dictionary with the following structure:
        {
            "sourceItemId": "1a201f2a-d1d8-45c0-8c61-1676338517de",
            "itemType": "SemanticModel"
        }
        If None, all items will be deployed.
    note : str, default=None
        An optional note to include with the deployment.
    allow_cross_region_deployment : bool, default=False
        Indicates whether cross region deployment is enabled. True - enabled, False - disabled. Default value is False.
    capacity : str | uuid.UUID, default=None
        The capacity name or ID to use for the deployment operation if creating a new workspace. Required when deploying to a stage that has no assigned workspaces, otherwise it is ignored. The deployment will fail if the new workspace configuration details aren't provided when required.
    workspace_name : str, default=None
        The workspace name to use for the deployment operation if creating a new workspace. Required when deploying to a stage that has no assigned workspaces, otherwise it is ignored. The deployment will fail if the new workspace configuration details aren't provided when required.
    """

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    payload = {
        "sourceStageId": source_stage_id,
        "targetStageId": target_stage_id,
    }
    if note:
        payload["note"] = note

    if items:
        if isinstance(items, dict):
            items = [items]

        if not isinstance(items, list):
            raise ValueError(
                f"{icons.red_dot} The 'items' parameter must be a list of dictionaries."
            )

        payload["items"] = items

    if allow_cross_region_deployment:
        payload["options"] = {"allowCrossRegionDeployment": True}

    if capacity and workspace_name:
        capacity_id = resolve_capacity_id(capacity)
        payload["createdWorkspaceDetails"] = {
            "capacityId": capacity_id,
            "name": workspace_name,
        }

    _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/deploy",
        method="post",
        payload=payload,
        status_codes=[200, 202],
        lro_return_status_code=True,
    )

    print(
        f"{icons.green_dot} The deployment from stage '{source_stage_id}' to stage '{target_stage_id}' in the '{deployment_pipeline}' deployment pipeline has been initiated successfully."
    )

    if capacity and workspace_name:
        print(
            f"{icons.info} A new workspace '{workspace_name}' will be created in the specified capacity for the deployment."
        )



from sempy._utils._log import log
import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    resolve_capacity_id,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
from uuid import UUID
import sempy_labs._icons as icons


@log
def list_gateways() -> pd.DataFrame:
    """
    Returns a list of all gateways the user has permission for, including on-premises, on-premises (personal mode), and virtual network gateways.

    This is a wrapper function for the following API: `Gateways - List Gateways <https://learn.microsoft.com/rest/api/fabric/core/gateways/list-gateways>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all gateways the user has permission for, including on-premises, on-premises (personal mode), and virtual network gateways.
    """

    columns = {
        "Gateway Name": "string",
        "Gateway Id": "string",
        "Type": "string",
        "Public Key Exponent": "string",
        "Public Key Modulus": "string",
        "Version": "string",
        "Number Of Member Gateways": "int",
        "Load Balancing Setting": "string",
        "Allow Cloud Connection Refresh": "bool",
        "Allow Custom Connectors": "bool",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/gateways", client="fabric_sp", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Gateway Name": v.get("displayName"),
                    "Gateway Id": v.get("id"),
                    "Type": v.get("type"),
                    "Public Key Exponent": v.get("publicKey", {}).get("exponent"),
                    "Public Key Modulus": v.get("publicKey", {}).get("modulus"),
                    "Version": v.get("version"),
                    "Number Of Member Gateways": v.get("numberOfMemberGateways", 0),
                    "Load Balancing Setting": v.get("loadBalancingSetting"),
                    "Allow Cloud Connection Refresh": v.get(
                        "allowCloudConnectionRefresh"
                    ),
                    "Allow Custom Connectors": v.get("allowCustomConnectors"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def resolve_gateway_id(gateway: str | UUID) -> UUID:

    if _is_valid_uuid(gateway):
        return gateway
    else:
        dfG = list_gateways()
        dfG_filt = dfG[dfG["Gateway Name"] == gateway]

        if dfG_filt.empty:
            raise ValueError(f"{icons.red_dot} The '{gateway}' gateway does not exist.")

        return dfG_filt["Gateway Id"].iloc[0]


@log
def delete_gateway(gateway: str | UUID):
    """
    Deletes a gateway.

    This is a wrapper function for the following API: `Gateways - Delete Gateway <https://learn.microsoft.com/rest/api/fabric/core/gateways/delete-gateway>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.
    """

    gateway_id = resolve_gateway_id(gateway)
    _base_api(request=f"/v1/gateways/{gateway_id}", client="fabric_sp", method="delete")
    print(f"{icons.green_dot} The '{gateway}' gateway has been deleted.")


@log
def list_gateway_role_assigments(gateway: str | UUID) -> pd.DataFrame:
    """
    Returns a list of gateway role assignments.

    This is a wrapper function for the following API: `Gateways - List Gateway Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/gateways/list-gateway-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of gateway role assignments.
    """

    columns = {
        "Gateway Role Assignment Id": "string",
        "Principal Id": "string",
        "Principal Type": "string",
        "Role": "string",
    }
    df = _create_dataframe(columns=columns)
    gateway_id = resolve_gateway_id(gateway)
    responses = _base_api(
        request=f"/v1/gateways/{gateway_id}/roleAssignments",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Gateway Role Assignment Id": v.get("id"),
                    "Principal Id": v.get("principal", {}).get("id"),
                    "Principal Type": v.get("principal", {}).get("type"),
                    "Role": v.get("role"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_gateway_role_assignment(gateway: str | UUID, role_assignment_id: UUID):
    """
    Delete the specified role assignment for the gateway.

    This is a wrapper function for the following API: `Gateways - Delete Gateway Role Assignment <https://learn.microsoft.com/rest/api/fabric/core/gateways/delete-gateway-role-assignment>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.
    role_assignment_id : uuid.UUID
        The role assignment ID.
    """

    gateway_id = resolve_gateway_id(gateway)
    _base_api(
        request=f"/v1/gateways/{gateway_id}/roleAssignments/{role_assignment_id}",
        client="fabric_sp",
        method="delete",
    )

    print(
        f"{icons.green_dot} The '{role_assignment_id}' role assignment for the '{gateway}' gateway has been deleted."
    )


@log
def resolve_gateway_member_id(gateway: str | UUID, gateway_member: str | UUID) -> UUID:

    gateway_id = resolve_gateway_id(gateway)
    dfM = list_gateway_members(gateway=gateway_id)

    if _is_valid_uuid(gateway_member):
        dfM_filt = dfM[dfM["Member Id"] == gateway_member]
    else:
        dfM_filt = dfM[dfM["Member Name"] == gateway_member]
    if len(dfM_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{gateway_member}' gateway member does not exist within the '{gateway}' gateway."
        )

    return dfM_filt["Member Id"].iloc[0]


@log
def delete_gateway_member(gateway: str | UUID, gateway_member: str | UUID):
    """
    Delete gateway member of an on-premises gateway.

    This is a wrapper function for the following API: `Gateways - Delete Gateway Member <https://learn.microsoft.com/rest/api/fabric/core/gateways/delete-gateway-member>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.
    gateway_member : str | uuid.UUID
        The gateway member name or ID.
    """

    gateway_id = resolve_gateway_id(gateway)
    member_id = resolve_gateway_member_id(
        gateway=gateway_id, gateway_member=gateway_member
    )

    _base_api(
        request=f"/v1/gateways/{gateway_id}/members/{member_id}",
        client="fabric_sp",
        method="delete",
    )
    print(
        f"{icons.green_dot} The '{member_id}' member for the '{gateway}' gateway has been deleted."
    )


@log
def list_gateway_members(gateway: str | UUID) -> pd.DataFrame:
    """
    Lists gateway members of an on-premises gateway.

    This is a wrapper function for the following API: `Gateways - List Gateway Members <https://learn.microsoft.com/rest/api/fabric/core/gateways/list-gateway-members>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of gateway members of an on-premises gateway.
    """

    gateway_id = resolve_gateway_id(gateway)

    columns = {
        "Member Id": "string",
        "Member Name": "string",
        "Public Key Exponent": "string",
        "Public Key Modulus": "string",
        "Version": "string",
        "Enabled": "bool",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1/gateways/{gateway_id}/members", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Member Id": v.get("id"),
                "Member Name": v.get("displayName"),
                "Public Key Exponent": v.get("publicKey", {}).get("exponent"),
                "Public Key Modulus": v.get("publicKey", {}).get("modulus"),
                "Version": v.get("version"),
                "Enabled": v.get("enabled"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def create_vnet_gateway(
    name: str,
    capacity: str | UUID,
    inactivity_minutes_before_sleep: int,
    number_of_member_gateways: int,
    subscription_id: str,
    resource_group: str,
    virtual_network: str,
    subnet: str,
):
    """
    Creates a virtual network gateway.

    This is a wrapper function for the following API: `Gateways - Create Gateway <https://learn.microsoft.com/rest/api/fabric/core/gateways/create-gateway>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The gateway name.
    capacity : str | uuid.UUID
        The capacity name or Id.
    inactivity_minutes_before_sleep : int
        The minutes of inactivity before the virtual network gateway goes into auto-sleep. Must be one of the following values: 30, 60, 90, 120, 150, 240, 360, 480, 720, 1440.
    number_of_member_gateways: int
        The number of member gateways. A number between 1 and 7.
    subscription_id : str
        The subscription ID.
    resource_group : str
        The name of the resource group.
    virtual_network : str
        The name of the virtual network.
    subnet : str
        The name of the subnet.
    """

    capacity_id = resolve_capacity_id(capacity=capacity)
    payload = {
        "type": "VirtualNetwork",
        "displayName": name,
        "capacityId": capacity_id,
        "virtualNetworkAzureResource": {
            "subscriptionId": subscription_id,
            "resourceGroupName": resource_group,
            "virtualNetworkName": virtual_network,
            "subnetName": subnet,
        },
        "inactivityMinutesBeforeSleep": inactivity_minutes_before_sleep,
        "numberOfMemberGateways": number_of_member_gateways,
    }

    _base_api(
        request="/v1/gateways",
        client="fabric_sp",
        method="post",
        payload=payload,
        status_codes=201,
    )

    print(
        f"{icons.green_dot} The '{name}' gateway was created within the '{capacity}' capacity."
    )


@log
def update_on_premises_gateway(
    gateway: str | UUID,
    allow_cloud_connection_refresh: Optional[bool] = None,
    allow_custom_connectors: Optional[bool] = None,
    load_balancing_setting: Optional[str] = None,
):
    """
    Updates an on-premises gateway.

    This is a wrapper function for the following API: `Gateways - Update Gateway <https://learn.microsoft.com/rest/api/fabric/core/gateways/update-gateway>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.
    allow_cloud_connection_refresh : bool, default=None
        Whether to allow cloud connections to refresh through this on-premises gateway. True - Allow, False - Do not allow.
    allow_custom_connectors : bool, default=None
        Whether to allow custom connectors to be used with this on-premises gateway. True - Allow, False - Do not allow.
    load_balancing_setting : str, default=None
        The `load balancing setting <https://learn.microsoft.com/rest/api/fabric/core/gateways/update-gateway?tabs=HTTP#loadbalancingsetting>`_ of the on-premises gateway.
    """

    gateway_id = resolve_gateway_id(gateway)

    payload = {}

    if allow_cloud_connection_refresh is not None:
        payload["allowCloudConnectionRefresh"] = allow_cloud_connection_refresh
    if allow_custom_connectors is not None:
        payload["allowCustomConnectors"] = allow_custom_connectors
    if load_balancing_setting is not None:
        payload["loadBalancingSetting"] = load_balancing_setting

    if not payload:
        raise ValueError(
            f"{icons.yellow_dot} The '{gateway}' gateway has not been update as no valid settings were provided."
        )

    payload["type"] = "OnPremises"

    _base_api(
        request=f"/v1/gateways/{gateway_id}",
        client="fabric_sp",
        method="patch",
        payload=payload,
    )

    print(f"{icons.green_dot} The '{gateway}' has been updated accordingly.")


@log
def update_vnet_gateway(
    gateway: str | UUID,
    capacity: str | UUID,
    inactivity_minutes_before_sleep: Optional[int] = None,
    number_of_member_gateways: Optional[int] = None,
):
    """
    Updates a virtual network gateway.

    This is a wrapper function for the following API: `Gateways - Update Gateway <https://learn.microsoft.com/rest/api/fabric/core/gateways/update-gateway>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    gateway : str | uuid.UUID
        The gateway name or ID.
    capacity: str | uuid.UUID
        The capacity name or ID.
    inactivity_minutes_before_sleep : int, default=None
        The minutes of inactivity before the virtual network gateway goes into auto-sleep. Must be one of the following values: 30, 60, 90, 120, 150, 240, 360, 480, 720, 1440.
    number_of_member_gateways : int, default=None
        The number of member gateways. A number between 1 and 7.
    """

    gateway_id = resolve_gateway_id(gateway)

    payload = {}

    if capacity is not None:
        capacity_id = resolve_capacity_id(capacity=capacity)
        payload["capacityId"] = capacity_id
    if inactivity_minutes_before_sleep is not None:
        payload["inactivityMinutesBeforeSleep"] = inactivity_minutes_before_sleep
    if number_of_member_gateways is not None:
        payload["numberOfMemberGateways"] = number_of_member_gateways

    if not payload:
        raise ValueError(
            f"{icons.yellow_dot} The '{gateway}' gateway has not been update as no valid settings were provided."
        )

    payload["type"] = "VirtualNetwork"

    _base_api(
        request=f"/v1/gateways/{gateway_id}",
        client="fabric_sp",
        method="patch",
        payload=payload,
    )
    print(f"{icons.green_dot} The '{gateway}' has been updated accordingly.")


@log
def bind_semantic_model_to_gateway(
    dataset: str | UUID,
    gateway: str | UUID,
    workspace: Optional[str | UUID] = None,
    data_source_object_ids: Optional[list[UUID]] = None,
):
    """
    Binds the specified dataset from the specified workspace to the specified gateway.

    This is a wrapper function for the following API: `Datasets - Bind To Gateway In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/bind-to-gateway-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or ID of the semantic model.
    gateway : str | uuid.UUID
        The name or ID of the gateway.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    data_source_object_ids : list[uuid.UUID], default=None
        A list of data source object IDs to bind to the gateway.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset, workspace=workspace_id
    )

    gateway_id = resolve_gateway_id(gateway)
    payload = {
        "gatewayObjectId": gateway_id,
    }
    if data_source_object_ids is not None:
        payload["datasourceObjectIds"] = data_source_object_ids

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/Default.BindToGateway",
        client="fabric_sp",
        method="post",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has been binded to the '{gateway_id}' gateway."
    )



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    get_model_id,
    _base_api,
)
from typing import Optional, Literal
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def approved_for_copilot(
    dataset: str | UUID,
    approved_for_copilot: bool,
    workspace: Optional[str | UUID] = None,
):
    """
    Turn on this setting to approve your model for Copilot use. This improves visibility in search results and removes the warning in standalone Copilot in Power BI that your organization hasn't approved the data.

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or UUID of the semantic model.
    approved_for_copilot : bool
        Set to True to approve the model for Copilot use, or False to disapprove it.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=workspace_id
    )
    payload = {"preppedForCopilot": approved_for_copilot, "isReadOnly": False}

    _base_api(
        request=f"metadata/workspace/{workspace_id}/model/{item_id}/preppedCopilot",
        client="internal",
        method="put",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{item_name}' semantic model within the '{workspace_name}' workspace has been set to to approved_for_copilot={approved_for_copilot}."
    )


@log
def set_endorsement(
    dataset: str | UUID,
    endorsement: Literal["None", "Promoted", "Certified", "Master data"],
    workspace: Optional[str | UUID] = None,
):
    """
    Sets the endorsement status for a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    endorsement : typing.Literal["None", "Promoted", "Certified", "Master data"]
        The endorsement status to set for the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=workspace_id
    )
    id = get_model_id(item_id=item_id)
    endorsement = endorsement.strip().lower()

    endorsement_mapping = {
        "none": 0,
        "promoted": 1,
        "certified": 2,
        "master data": 3,
    }

    if endorsement not in endorsement_mapping:
        raise ValueError(
            "Endorsement must be either 'None', 'Promoted', 'Certified', or 'Master data'."
        )

    stage = endorsement_mapping.get(endorsement)
    payload = {"stage": stage}

    _base_api(
        request=f"metadata/gallery/models/{id}",
        client="internal",
        method="put",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The endorsement for the '{item_name}' semantic model within the '{workspace_name}' workspace has been set to '{endorsement}'."
    )


@log
def make_discoverable(
    dataset: str | UUID, make_discoverable: bool, workspace: Optional[str | UUID] = None
):
    """
    Allow users without access to this semantic model to discover it and request permissions to access the data. The semantic model must be endorsed as 'Promoted', 'Certified', or 'Master data' to be made discoverable.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    make_discoverable : bool
        Whether to make the semantic model discoverable.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=workspace_id
    )
    id = get_model_id(item_id=item_id)

    if make_discoverable:
        _base_api(
            request=f"metadata/gallery/discover/models/{id}",
            client="internal",
            method="post",
        )
        print(
            f"{icons.green_dot} The '{item_name}' semantic model within the '{workspace_name}' workspace is now discoverable in the gallery."
        )
    else:
        _base_api(
            request=f"metadata/gallery/discover/models/{id}",
            client="internal",
            method="delete",
        )
        print(
            f"{icons.green_dot} The '{item_name}' semantic model within the '{workspace_name}' workspace is no longer discoverable in the gallery."
        )



from ._items import (
    list_warehouse_snapshots,
    delete_warehouse_snapshot,
    create_warehouse_snapshot,
    update_warehouse_snapshot,
)

__all__ = [
    "list_warehouse_snapshots",
    "delete_warehouse_snapshot",
    "create_warehouse_snapshot",
    "update_warehouse_snapshot",
]



import sempy
import pandas as pd
from uuid import UUID
from IPython.display import display
from typing import Optional
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
)
from sempy_labs.lakehouse._get_lakehouse_columns import get_lakehouse_columns
import sempy_labs._icons as icons
from sempy_labs.directlake._sources import get_direct_lake_sources
from sempy_labs.tom import connect_semantic_model


@log
def direct_lake_schema_compare(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Checks that all the tables in a Direct Lake semantic model map to tables in their corresponding lakehouse and that the columns in each table exist.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    sources = get_direct_lake_sources(dataset=dataset_id, workspace=workspace_id)[0]

    source_type = sources.get("itemType")

    if source_type != "Lakehouse":
        raise ValueError(
            f"{icons.red_dot} The source of the '{dataset_name}' semantic model within the '{workspace_name}' workspace is not a Fabric lakehouse. This function only supports Direct Lake semantic models which source from Fabric lakehouses."
        )

    lakehouse_name = sources.get("itemName")
    lakehouse_id = sources.get("itemId")
    lakehouse_workspace_id = sources.get("workspaceId")
    lakehouse_workspace_name = sources.get("workspaceName")
    df_lake = get_lakehouse_columns(
        lakehouse=lakehouse_id, workspace=lakehouse_workspace_id
    )

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    with connect_semantic_model(dataset=dataset_id, workspace=workspace_id) as tom:
        if tom.is_direct_lake is False:
            raise ValueError(
                f"{icons.red_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake mode."
            )
        missing_tables = []
        missing_columns = []
        for p in tom.all_partitions():
            if p.Mode == TOM.ModeType.DirectLake:
                table_name = p.Parent.Name
                source_table = p.Source.EntityName
                df_lake_filt = df_lake[df_lake["Table Name"] == source_table]
                if df_lake_filt.empty:
                    missing_tables.append(
                        {
                            "Table Name": table_name,
                            "Source Table": source_table,
                        }
                    )
                else:
                    for c in p.Parent.Columns:
                        col_name = c.Name
                        source_col = c.SourceColumn
                        df_lake_filt = df_lake[
                            (df_lake["Table Name"] == source_table)
                            & (df_lake["Column Name"] == source_col)
                        ]
                        if df_lake_filt.empty:
                            missing_columns.append(
                                {
                                    "Table Name": table_name,
                                    "Source Table": source_table,
                                    "Column Name": col_name,
                                    "Source Column": source_col,
                                }
                            )

        if len(missing_tables) == 0:
            print(
                f"{icons.green_dot} All tables exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace_name}' workspace."
            )
        else:
            print(
                f"{icons.yellow_dot} The following tables exist in the '{dataset_name}' semantic model within the '{workspace_name}' workspace"
                f" but do not exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace_name}' workspace."
            )
            display(pd.DataFrame(missing_tables))
        if len(missing_columns) == 0:
            print(
                f"{icons.green_dot} All columns exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace_name}' workspace."
            )
        else:
            print(
                f"{icons.yellow_dot} The following columns exist in the '{dataset_name}' semantic model within the '{workspace_name}' workspace "
                f"but do not exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace_name}' workspace."
            )
            display(pd.DataFrame(missing_columns))



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    get_model_id,
    _base_api,
)
from typing import Optional
import sempy_labs._icons as icons
from sempy._utils._log import log
from uuid import UUID


@log
def set_autosync(
    dataset: str | UUID, workspace: Optional[str | UUID] = None, enable: bool = True
):
    """
    Enables or disables AutoSync for a Direct Lake semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    enable : bool, default=True
        Whether to enable (True) or disable (False) AutoSync.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    id = get_model_id(item_id=dataset_id)

    payload = {"directLakeAutoSync": enable}
    _base_api(
        request=f"metadata/models/{id}/settings",
        client="internal",
        method="post",
        payload=payload,
        status_codes=204,
    )

    print(
        f"{icons.green_dot} Direct Lake AutoSync has been {'enabled' if enable else 'disabled'} for the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )



import sempy
import pandas as pd
import re
from typing import Optional
from sempy._utils._log import log


@log
def model_bpa_rules(
    dependencies: Optional[pd.DataFrame] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Shows the default rules for the semantic model BPA used by the run_model_bpa function.

    Parameters
    ----------
    dependencies : pd.DataFrame, default=None
        A pandas dataframe with the output of the 'get_model_calc_dependencies' function.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe containing the default rules for the run_model_bpa function.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    if "dataset" in kwargs:
        print(
            "The 'dataset' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["dataset"]
    if "workspace" in kwargs:
        print(
            "The 'workspace' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["workspace"]

    rules = pd.DataFrame(
        [
            (
                "Performance",
                "Column",
                "Warning",
                "Do not use floating point data types",
                lambda obj, tom: obj.DataType == TOM.DataType.Double,
                'The "Double" floating point data type should be avoided, as it can result in unpredictable roundoff errors and decreased performance in certain scenarios. Use "Int64" or "Decimal" where appropriate (but note that "Decimal" is limited to 4 digits after the decimal sign).',
            ),
            (
                "Performance",
                "Column",
                "Warning",
                "Avoid using calculated columns",
                lambda obj, tom: obj.Type == TOM.ColumnType.Calculated,
                "Calculated columns do not compress as well as data columns so they take up more memory. They also slow down processing times for both the table as well as process recalc. Offload calculated column logic to your data warehouse and turn these calculated columns into data columns.",
                "https://www.elegantbi.com/post/top10bestpractices",
            ),
            (
                "Performance",
                "Relationship",
                "Warning",
                "Check if bi-directional and many-to-many relationships are valid",
                lambda obj, tom: (
                    obj.FromCardinality == TOM.RelationshipEndCardinality.Many
                    and obj.ToCardinality == TOM.RelationshipEndCardinality.Many
                )
                or str(obj.CrossFilteringBehavior) == "BothDirections",
                "Bi-directional and many-to-many relationships may cause performance degradation or even have unintended consequences. Make sure to check these specific relationships to ensure they are working as designed and are actually necessary.",
                "https://www.sqlbi.com/articles/bidirectional-relationships-and-ambiguity-in-dax",
            ),
            (
                "Performance",
                "Row Level Security",
                "Info",
                "Check if dynamic row level security (RLS) is necessary",
                lambda obj, tom: any(
                    re.search(pattern, obj.FilterExpression, flags=re.IGNORECASE)
                    for pattern in ["USERPRINCIPALNAME()", "USERNAME()"]
                ),
                "Usage of dynamic row level security (RLS) can add memory and performance overhead. Please research the pros/cons of using it.",
                "https://docs.microsoft.com/power-bi/admin/service-admin-rls",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Avoid using many-to-many relationships on tables used for dynamic row level security",
                lambda obj, tom: any(
                    r.FromCardinality == TOM.RelationshipEndCardinality.Many
                    and r.ToCardinality == TOM.RelationshipEndCardinality.Many
                    for r in tom.used_in_relationships(object=obj)
                )
                and any(t.Name == obj.Name for t in tom.all_rls()),
                "Using many-to-many relationships on tables which use dynamic row level security can cause serious query performance degradation. This pattern's performance problems compound when snowflaking multiple many-to-many relationships against a table which contains row level security. Instead, use one of the patterns shown in the article below where a single dimension table relates many-to-one to a security table.",
                "https://www.elegantbi.com/post/dynamicrlspatterns",
            ),
            (
                "Performance",
                "Relationship",
                "Warning",
                "Many-to-many relationships should be single-direction",
                lambda obj, tom: (
                    obj.FromCardinality == TOM.RelationshipEndCardinality.Many
                    and obj.ToCardinality == TOM.RelationshipEndCardinality.Many
                )
                and obj.CrossFilteringBehavior
                == TOM.CrossFilteringBehavior.BothDirections,
            ),
            (
                "Performance",
                "Column",
                "Warning",
                "Set IsAvailableInMdx to false on non-attribute columns",
                lambda obj, tom: tom.is_direct_lake() is False
                and obj.IsAvailableInMDX
                and (obj.IsHidden or obj.Parent.IsHidden)
                and obj.SortByColumn is None
                and not any(tom.used_in_sort_by(column=obj))
                and not any(tom.used_in_hierarchies(column=obj)),
                "To speed up processing time and conserve memory after processing, attribute hierarchies should not be built for columns that are never used for slicing by MDX clients. In other words, all hidden columns that are not used as a Sort By Column or referenced in user hierarchies should have their IsAvailableInMdx property set to false. The IsAvailableInMdx property is not relevant for Direct Lake models.",
                "https://blog.crossjoin.co.uk/2018/07/02/isavailableinmdx-ssas-tabular",
            ),
            (
                "Performance",
                "Partition",
                "Warning",
                "Set 'Data Coverage Definition' property on the DirectQuery partition of a hybrid table",
                lambda obj, tom: tom.is_hybrid_table(table_name=obj.Parent.Name)
                and obj.Mode == TOM.ModeType.DirectQuery
                and obj.DataCoverageDefinition is None,
                "Setting the 'Data Coverage Definition' property may lead to better performance because the engine knows when it can only query the import-portion of the table and when it needs to query the DirectQuery portion of the table.",
                "https://learn.microsoft.com/analysis-services/tom/table-partitions?view=asallproducts-allversions",
            ),
            (
                "Performance",
                "Model",
                "Warning",
                "Dual mode is only relevant for dimension tables if DirectQuery is used for the corresponding fact table",
                lambda obj, tom: not any(
                    p.Mode == TOM.ModeType.DirectQuery for p in tom.all_partitions()
                )
                and any(p.Mode == TOM.ModeType.Dual for p in tom.all_partitions()),
                "Only use Dual mode for dimension tables/partitions where a corresponding fact table is in DirectQuery. Using Dual mode in other circumstances (i.e. rest of the model is in Import mode) may lead to performance issues especially if the number of measures in the model is high.",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Set dimensions tables to dual mode instead of import when using DirectQuery on fact tables",
                lambda obj, tom: sum(
                    1 for p in obj.Partitions if p.Mode == TOM.ModeType.Import
                )
                == 1
                and obj.Partitions.Count == 1
                and tom.has_hybrid_table()
                and any(
                    r.ToCardinality == TOM.RelationshipEndCardinality.One
                    and r.ToTable.Name == obj.Name
                    for r in tom.used_in_relationships(object=obj)
                ),
                "When using DirectQuery, dimension tables should be set to Dual mode in order to improve query performance.",
                "https://learn.microsoft.com/power-bi/transform-model/desktop-storage-mode#propagation-of-the-dual-setting",
            ),
            (
                "Performance",
                "Partition",
                "Warning",
                "Minimize Power Query transformations",
                lambda obj, tom: obj.SourceType == TOM.PartitionSourceType.M
                and any(
                    item in obj.Source.Expression
                    for item in [
                        'Table.Combine("',
                        'Table.Join("',
                        'Table.NestedJoin("',
                        'Table.AddColumn("',
                        'Table.Group("',
                        'Table.Sort("',
                        'Table.Pivot("',
                        'Table.Unpivot("',
                        'Table.UnpivotOtherColumns("',
                        'Table.Distinct("',
                        '[Query=(""SELECT',
                        "Value.NativeQuery",
                        "OleDb.Query",
                        "Odbc.Query",
                    ]
                ),
                "Minimize Power Query transformations in order to improve model processing performance. It is a best practice to offload these transformations to the data warehouse if possible. Also, please check whether query folding is occurring within your model. Please reference the article below for more information on query folding.",
                "https://docs.microsoft.com/power-query/power-query-folding",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Consider a star-schema instead of a snowflake architecture",
                lambda obj, tom: obj.CalculationGroup is None
                and (
                    any(
                        r.FromTable.Name == obj.Name
                        for r in tom.used_in_relationships(object=obj)
                    )
                    and any(
                        r.ToTable.Name == obj.Name
                        for r in tom.used_in_relationships(object=obj)
                    )
                ),
                "Generally speaking, a star-schema is the optimal architecture for tabular models. That being the case, there are valid cases to use a snowflake approach. Please check your model and consider moving to a star-schema architecture.",
                "https://docs.microsoft.com/power-bi/guidance/star-schema",
            ),
            (
                "Performance",
                "Model",
                "Warning",
                "Avoid using views when using Direct Lake mode",
                lambda obj, tom: tom.is_direct_lake_using_view(),
                "In Direct Lake mode, views will always fall back to DirectQuery. Thus, in order to obtain the best performance use lakehouse tables instead of views.",
                "https://learn.microsoft.com/fabric/get-started/direct-lake-overview#fallback",
            ),
            (
                "Performance",
                "Measure",
                "Warning",
                "Avoid adding 0 to a measure",
                lambda obj, tom: obj.Expression.replace(" ", "").startswith("0+")
                or obj.Expression.replace(" ", "").endswith("+0")
                or re.search(
                    r"DIVIDE\s*\(\s*[^,]+,\s*[^,]+,\s*0\s*\)",
                    obj.Expression,
                    flags=re.IGNORECASE,
                )
                or re.search(
                    r"IFERROR\s*\(\s*[^,]+,\s*0\s*\)",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "Adding 0 to a measure in order for it not to show a blank value may negatively impact performance.",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Reduce usage of calculated tables",
                lambda obj, tom: tom.is_field_parameter(table_name=obj.Name) is False
                and tom.is_calculated_table(table_name=obj.Name),
                "Migrate calculated table logic to your data warehouse. Reliance on calculated tables will lead to technical debt and potential misalignments if you have multiple models on your platform.",
            ),
            (
                "Performance",
                "Column",
                "Warning",
                "Reduce usage of calculated columns that use the RELATED function",
                lambda obj, tom: obj.Type == TOM.ColumnType.Calculated
                and re.search(r"related\s*\(", obj.Expression, flags=re.IGNORECASE),
                "Calculated columns do not compress as well as data columns and may cause longer processing times. As such, calculated columns should be avoided if possible. One scenario where they may be easier to avoid is if they use the RELATED function.",
                "https://www.sqlbi.com/articles/storage-differences-between-calculated-columns-and-calculated-tables",
            ),
            (
                "Performance",
                "Model",
                "Warning",
                "Avoid excessive bi-directional or many-to-many relationships",
                lambda obj, tom: (
                    (
                        sum(
                            1
                            for r in obj.Relationships
                            if r.CrossFilteringBehavior
                            == TOM.CrossFilteringBehavior.BothDirections
                        )
                        + sum(
                            1
                            for r in obj.Relationships
                            if (
                                r.FromCardinality == TOM.RelationshipEndCardinality.Many
                            )
                            and (r.ToCardinality == TOM.RelationshipEndCardinality.Many)
                        )
                    )
                    / max(int(obj.Relationships.Count), 1)
                )
                > 0.3,
                "Limit use of b-di and many-to-many relationships. This rule flags the model if more than 30% of relationships are bi-di or many-to-many.",
                "https://www.sqlbi.com/articles/bidirectional-relationships-and-ambiguity-in-dax",
            ),
            # ('Performance', 'Column', 'Warning', 'Avoid bi-directional or many-to-many relationships against high-cardinality columns',
            # lambda obj, tom: ((str(r.FromCardinality) == 'Many' and str(r.ToCardinality == 'Many'))  or (str(r.CrossFilteringBehavior) == 'BothDirections') for r in tom.used_in_relationships(object = obj)) and tom.cardinality(column = obj) > 100000,
            # 'For best performance, it is recommended to avoid using bi-directional relationships against high-cardinality columns',
            # ),
            (
                "Performance",
                "Table",
                "Warning",
                "Remove auto-date table",
                lambda obj, tom: tom.is_calculated_table(table_name=obj.Name)
                and (
                    obj.Name.startswith("DateTableTemplate_")
                    or obj.Name.startswith("LocalDateTable_")
                ),
                "Avoid using auto-date tables. Make sure to turn off auto-date table in the settings in Power BI Desktop. This will save memory resources.",
                "https://www.youtube.com/watch?v=xu3uDEHtCrg",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Date/calendar tables should be marked as a date table",
                lambda obj, tom: (
                    re.search(r"date", obj.Name, flags=re.IGNORECASE)
                    or re.search(r"calendar", obj.Name, flags=re.IGNORECASE)
                )
                and str(obj.DataCategory) != "Time",
                "This rule looks for tables that contain the words 'date' or 'calendar' as they should likely be marked as a date table.",
                "https://docs.microsoft.com/power-bi/transform-model/desktop-date-tables",
            ),
            (
                "Performance",
                "Table",
                "Warning",
                "Large tables should be partitioned",
                lambda obj, tom: tom.is_direct_lake() is False
                and int(obj.Partitions.Count) == 1
                and tom.row_count(object=obj) > 25000000,
                "Large tables should be partitioned in order to optimize processing. This is not relevant for semantic models in Direct Lake mode as they can only have one partition per table.",
            ),
            (
                "Performance",
                "Row Level Security",
                "Warning",
                "Limit row level security (RLS) logic",
                lambda obj, tom: any(
                    item in obj.FilterExpression.lower()
                    for item in [
                        "right(",
                        "left(",
                        "filter(",
                        "upper(",
                        "lower(",
                        "find(",
                    ]
                ),
                "Try to simplify the DAX used for row level security. Usage of the functions within this rule can likely be offloaded to the upstream systems (data warehouse).",
            ),
            (
                "Performance",
                "Model",
                "Warning",
                "Model should have a date table",
                lambda obj, tom: not any(
                    (c.IsKey and c.DataType == TOM.DataType.DateTime)
                    and str(t.DataCategory) == "Time"
                    for t in obj.Tables
                    for c in t.Columns
                ),
                "Generally speaking, models should generally have a date table. Models that do not have a date table generally are not taking advantage of features such as time intelligence or may not have a properly structured architecture.",
            ),
            # ('Performance', 'Measure', 'Warning', 'Measures using time intelligence and model is using Direct Query',
            # lambda obj, tom: any(str(p.Mode) == 'DirectQuery' for p in tom.all_partitions()) and any(re.search(pattern + '\s*\(', obj.Expression, flags=re.IGNORECASE) for pattern in ['CLOSINGBALANCEMONTH', 'CLOSINGBALANCEQUARTER', 'CLOSINGBALANCEYEAR', \
            #    'DATEADD', 'DATESBETWEEN', 'DATESINPERIOD', 'DATESMTD', 'DATESQTD', 'DATESYTD', 'ENDOFMONTH', 'ENDOFQUARTER', 'ENDOFYEAR', 'FIRSTDATE', 'FIRSTNONBLANK', 'FIRSTNONBLANKVALUE', 'LASTDATE', 'LASTNONBLANK', 'LASTNONBLANKVALUE', \
            #    'NEXTDAY', 'NEXTMONTH', 'NEXTQUARTER', 'NEXTYEAR', 'OPENINGBALANCEMONTH', 'OPENINGBALANCEQUARTER', 'OPENINGBALANCEYEAR', 'PARALLELPERIOD', 'PREVIOUSDAY', 'PREVIOUSMONTH', 'PREVIOUSQUARTER', 'PREVIOUSYEAR', 'SAMEPERIODLASTYEAR', \
            #    'STARTOFMONTH', 'STARTOFQUARTER', 'STARTOFYEAR', 'TOTALMTD', 'TOTALQTD', 'TOTALYTD']),
            # 'At present, time intelligence functions are known to not perform as well when using Direct Query. If you are having performance issues, you may want to try alternative solutions such as adding columns in the fact table that show previous year or previous month data.',
            # ),
            (
                "Error Prevention",
                "Calculation Item",
                "Error",
                "Calculation items must have an expression",
                lambda obj, tom: len(obj.Expression) == 0,
                "Calculation items must have an expression. Without an expression, they will not show any values.",
            ),
            # ('Error Prevention', ['Table', 'Column', 'Measure', 'Hierarchy', 'Partition'], 'Error', 'Avoid invalid characters in names',
            # lambda obj, tom: obj.Name
            # 'This rule identifies if a name for a given object in your model (i.e. table/column/measure) which contains an invalid character. Invalid characters will cause an error when deploying the model (and failure to deploy). This rule has a fix expression which converts the invalid character into a space, resolving the issue.',
            # ),
            # ('Error Prevention', ['Table', 'Column', 'Measure', 'Hierarchy'], 'Error', 'Avoid invalid characters in descriptions',
            # lambda obj, tom: obj.Description
            # 'This rule identifies if a description for a given object in your model (i.e. table/column/measure) which contains an invalid character. Invalid characters will cause an error when deploying the model (and failure to deploy). This rule has a fix expression which converts the invalid character into a space, resolving the issue.',
            # ),
            (
                "Error Prevention",
                "Relationship",
                "Warning",
                "Relationship columns should be of the same data type",
                lambda obj, tom: obj.FromColumn.DataType != obj.ToColumn.DataType,
                "Columns used in a relationship should be of the same data type. Ideally, they will be of integer data type (see the related rule '[Formatting] Relationship columns should be of integer data type'). Having columns within a relationship which are of different data types may lead to various issues.",
            ),
            (
                "Error Prevention",
                "Column",
                "Error",
                "Data columns must have a source column",
                lambda obj, tom: obj.Type == TOM.ColumnType.Data
                and len(obj.SourceColumn) == 0,
                "Data columns must have a source column. A data column without a source column will cause an error when processing the model.",
            ),
            (
                "Error Prevention",
                "Column",
                "Warning",
                "Set IsAvailableInMdx to true on necessary columns",
                lambda obj, tom: tom.is_direct_lake() is False
                and obj.IsAvailableInMDX is False
                and (
                    any(tom.used_in_sort_by(column=obj))
                    or any(tom.used_in_hierarchies(column=obj))
                    or obj.SortByColumn is not None
                ),
                "In order to avoid errors, ensure that attribute hierarchies are enabled if a column is used for sorting another column, used in a hierarchy, used in variations, or is sorted by another column. The IsAvailableInMdx property is not relevant for Direct Lake models.",
            ),
            (
                "Error Prevention",
                "Table",
                "Error",
                "Avoid the USERELATIONSHIP function and RLS against the same table",
                lambda obj, tom: any(
                    re.search(
                        r"USERELATIONSHIP\s*\(\s*.+?(?=])\]\s*,\s*'*"
                        + re.escape(obj.Name)
                        + r"'*\[",
                        m.Expression,
                        flags=re.IGNORECASE,
                    )
                    for m in tom.all_measures()
                )
                and any(r.Table.Name == obj.Name for r in tom.all_rls()),
                "The USERELATIONSHIP function may not be used against a table which also leverages row-level security (RLS). This will generate an error when using the particular measure in a visual. This rule will highlight the table which is used in a measure's USERELATIONSHIP function as well as RLS.",
                "https://blog.crossjoin.co.uk/2013/05/10/userelationship-and-tabular-row-security",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Avoid using the IFERROR function",
                lambda obj, tom: re.search(
                    r"iferror\s*\(", obj.Expression, flags=re.IGNORECASE
                ),
                "Avoid using the IFERROR function as it may cause performance degradation. If you are concerned about a divide-by-zero error, use the DIVIDE function as it naturally resolves such errors as blank (or you can customize what should be shown in case of such an error).",
                "https://www.elegantbi.com/post/top10bestpractices",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Use the TREATAS function instead of INTERSECT for virtual relationships",
                lambda obj, tom: re.search(
                    r"intersect\s*\(", obj.Expression, flags=re.IGNORECASE
                ),
                "The TREATAS function is more efficient and provides better performance than the INTERSECT function when used in virutal relationships.",
                "https://www.sqlbi.com/articles/propagate-filters-using-treatas-in-dax",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "The EVALUATEANDLOG function should not be used in production models",
                lambda obj, tom: re.search(
                    r"evaluateandlog\s*\(",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "The EVALUATEANDLOG function is meant to be used only in development/test environments and should not be used in production models.",
                "https://pbidax.wordpress.com/2022/08/16/introduce-the-dax-evaluateandlog-function",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Measures should not be direct references of other measures",
                lambda obj, tom: any(
                    obj.Expression == f"[{m.Name}]" for m in tom.all_measures()
                ),
                "This rule identifies measures which are simply a reference to another measure. As an example, consider a model with two measures: [MeasureA] and [MeasureB]. This rule would be triggered for MeasureB if MeasureB's DAX was MeasureB:=[MeasureA]. Such duplicative measures should be removed.",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "No two measures should have the same definition",
                lambda obj, tom: any(
                    re.sub(r"\s+", "", obj.Expression)
                    == re.sub(r"\s+", "", m.Expression)
                    and obj.Name != m.Name
                    for m in tom.all_measures()
                ),
                "Two measures with different names and defined by the same DAX expression should be avoided to reduce redundancy.",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Avoid addition or subtraction of constant values to results of divisions",
                lambda obj, tom: re.search(
                    r"DIVIDE\s*\((\s*.*?)\)\s*[+-]\s*1|\/\s*.*(?=[-+]\s*1)",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "Adding a constant value may lead to performance degradation.",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Avoid using '1-(x/y)' syntax",
                lambda obj, tom: re.search(
                    r"[0-9]+\s*[-+]\s*[\(]*\s*SUM\s*\(\s*\'*[A-Za-z0-9 _]+\'*\s*\[[A-Za-z0-9 _]+\]\s*\)\s*/",
                    obj.Expression,
                    flags=re.IGNORECASE,
                )
                or re.search(
                    r"[0-9]+\s*[-+]\s*DIVIDE\s*\(",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "Instead of using the '1-(x/y)' or '1+(x/y)' syntax to achieve a percentage calculation, use the basic DAX functions (as shown below). Using the improved syntax will generally improve the performance. The '1+/-...' syntax always returns a value whereas the solution without the '1+/-...' does not (as the value may be 'blank'). Therefore the '1+/-...' syntax may return more rows/columns which may result in a slower query speed.    Let's clarify with an example:    Avoid this: 1 - SUM ( 'Sales'[CostAmount] ) / SUM( 'Sales'[SalesAmount] )  Better: DIVIDE ( SUM ( 'Sales'[SalesAmount] ) - SUM ( 'Sales'[CostAmount] ), SUM ( 'Sales'[SalesAmount] ) )  Best: VAR x = SUM ( 'Sales'[SalesAmount] ) RETURN DIVIDE ( x - SUM ( 'Sales'[CostAmount] ), x )",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Filter measure values by columns, not tables",
                lambda obj, tom: re.search(
                    r"CALCULATE\s*\(\s*[^,]+,\s*FILTER\s*\(\s*\'*[A-Za-z0-9 _]+\'*\s*,\s*\[[^\]]+\]",
                    obj.Expression,
                    flags=re.IGNORECASE,
                )
                or re.search(
                    r"CALCULATETABLE\s*\(\s*[^,]*,\s*FILTER\s*\(\s*\'*[A-Za-z0-9 _]+\'*\s*,\s*\[",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "Instead of using this pattern FILTER('Table',[Measure]>Value) for the filter parameters of a CALCULATE or CALCULATETABLE function, use one of the options below (if possible). Filtering on a specific column will produce a smaller table for the engine to process, thereby enabling faster performance. Using the VALUES function or the ALL function depends on the desired measure result.\nOption 1: FILTER(VALUES('Table'[Column]),[Measure] > Value)\nOption 2: FILTER(ALL('Table'[Column]),[Measure] > Value)",
                "https://docs.microsoft.com/power-bi/guidance/dax-avoid-avoid-filter-as-filter-argument",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Filter column values with proper syntax",
                lambda obj, tom: re.search(
                    r"CALCULATE\s*\(\s*[^,]+,\s*FILTER\s*\(\s*'*[A-Za-z0-9 _]+'*\s*,\s*'*[A-Za-z0-9 _]+'*\[[A-Za-z0-9 _]+\]",
                    obj.Expression,
                    flags=re.IGNORECASE,
                )
                or re.search(
                    r"CALCULATETABLE\s*\([^,]*,\s*FILTER\s*\(\s*'*[A-Za-z0-9 _]+'*\s*,\s*'*[A-Za-z0-9 _]+'*\[[A-Za-z0-9 _]+\]",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                "Instead of using this pattern FILTER('Table','Table'[Column]=\"Value\") for the filter parameters of a CALCULATE or CALCULATETABLE function, use one of the options below. As far as whether to use the KEEPFILTERS function, see the second reference link below.\nOption 1: KEEPFILTERS('Table'[Column]=\"Value\")\nOption 2: 'Table'[Column]=\"Value\"",
                "https://docs.microsoft.com/power-bi/guidance/dax-avoid-avoid-filter-as-filter-argument  Reference: https://www.sqlbi.com/articles/using-keepfilters-in-dax",
            ),
            (
                "DAX Expressions",
                "Measure",
                "Warning",
                "Use the DIVIDE function for division",
                lambda obj, tom: re.search(
                    r"\]\s*\/(?!\/)(?!\*)|\)\s*\/(?!\/)(?!\*)",
                    obj.Expression,
                    flags=re.IGNORECASE,
                ),
                'Use the DIVIDE  function instead of using "/". The DIVIDE function resolves divide-by-zero cases. As such, it is recommended to use to avoid errors.',
                "https://docs.microsoft.com/power-bi/guidance/dax-divide-function-operator",
            ),
            (
                "DAX Expressions",
                [
                    "Measure",
                    "Calculated Table",
                    "Calculated Column",
                    "Calculation Item",
                ],
                "Error",
                "Column references should be fully qualified",
                lambda obj, tom: any(
                    tom.unqualified_columns(object=obj, dependencies=dependencies)
                ),
                "Using fully qualified column references makes it easier to distinguish between column and measure references, and also helps avoid certain errors. When referencing a column in DAX, first specify the table name, then specify the column name in square brackets.",
                "https://www.elegantbi.com/post/top10bestpractices",
            ),
            (
                "DAX Expressions",
                [
                    "Measure",
                    "Calculated Table",
                    "Calculated Column",
                    "Calculation Item",
                ],
                "Error",
                "Measure references should be unqualified",
                lambda obj, tom: any(
                    tom.fully_qualified_measures(object=obj, dependencies=dependencies)
                ),
                "Using unqualified measure references makes it easier to distinguish between column and measure references, and also helps avoid certain errors. When referencing a measure using DAX, do not specify the table name. Use only the measure name in square brackets.",
                "https://www.elegantbi.com/post/top10bestpractices",
            ),
            (
                "DAX Expressions",
                "Relationship",
                "Warning",
                "Inactive relationships that are never activated",
                lambda obj, tom: obj.IsActive is False
                and not any(
                    re.search(
                        r"USERELATIONSHIP\s*\(\s*\'*"
                        + re.escape(obj.FromTable.Name)
                        + r"'*\["
                        + re.escape(obj.FromColumn.Name)
                        + r"\]\s*,\s*'*"
                        + re.escape(obj.ToTable.Name)
                        + r"'*\["
                        + re.escape(obj.ToColumn.Name)
                        + r"\]",
                        m.Expression,
                        flags=re.IGNORECASE,
                    )
                    for m in tom.all_measures()
                ),
                "Inactive relationships are activated using the USERELATIONSHIP function. If an inactive relationship is not referenced in any measure via this function, the relationship will not be used. It should be determined whether the relationship is not necessary or to activate the relationship via this method.",
                "https://dax.guide/userelationship",
            ),
            (
                "Maintenance",
                "Column",
                "Warning",
                "Remove unnecessary columns",
                lambda obj, tom: (obj.IsHidden or obj.Parent.IsHidden)
                and not any(tom.used_in_relationships(object=obj))
                and not any(tom.used_in_hierarchies(column=obj))
                and not any(tom.used_in_sort_by(column=obj))
                and any(tom.depends_on(object=obj, dependencies=dependencies)),
                "Hidden columns that are not referenced by any DAX expressions, relationships, hierarchy levels or Sort By-properties should be removed.",
            ),
            (
                "Maintenance",
                "Measure",
                "Warning",
                "Remove unnecessary measures",
                lambda obj, tom: obj.IsHidden
                and not any(tom.referenced_by(object=obj, dependencies=dependencies)),
                "Hidden measures that are not referenced by any DAX expressions should be removed for maintainability.",
            ),
            (
                "Maintenance",
                "Table",
                "Warning",
                "Ensure tables have relationships",
                lambda obj, tom: any(tom.used_in_relationships(object=obj)) is False
                and obj.CalculationGroup is None,
                "This rule highlights tables which are not connected to any other table in the model with a relationship.",
            ),
            (
                "Maintenance",
                "Table",
                "Warning",
                "Calculation groups with no calculation items",
                lambda obj, tom: obj.CalculationGroup is not None
                and not any(obj.CalculationGroup.CalculationItems),
                "Calculation groups have no function unless they have calculation items.",
            ),
            (
                "Maintenance",
                ["Column", "Measure", "Table"],
                "Info",
                "Visible objects with no description",
                lambda obj, tom: obj.IsHidden is False and len(obj.Description) == 0,
                "Add descriptions to objects. These descriptions are shown on hover within the Field List in Power BI Desktop. Additionally, you can leverage these descriptions to create an automated data dictionary.",
            ),
            (
                "Formatting",
                "Column",
                "Warning",
                "Provide format string for 'Date' columns",
                lambda obj, tom: (re.search(r"date", obj.Name, flags=re.IGNORECASE))
                and (obj.DataType == TOM.DataType.DateTime)
                and (
                    obj.FormatString.lower()
                    not in [
                        "mm/dd/yyyy",
                        "mm-dd-yyyy",
                        "dd/mm/yyyy",
                        "dd-mm-yyyy",
                        "yyyy-mm-dd",
                        "yyyy/mm/dd",
                    ]
                ),
                'Columns of type "DateTime" that have "Date" in their names should be formatted.',
            ),
            (
                "Formatting",
                "Column",
                "Warning",
                "Do not summarize numeric columns",
                lambda obj, tom: (
                    (obj.DataType == TOM.DataType.Int64)
                    or (obj.DataType == TOM.DataType.Decimal)
                    or (obj.DataType == TOM.DataType.Double)
                )
                and (str(obj.SummarizeBy) != "None")
                and not ((obj.IsHidden) or (obj.Parent.IsHidden)),
                'Numeric columns (integer, decimal, double) should have their SummarizeBy property set to "None" to avoid accidental summation in Power BI (create measures instead).',
            ),
            (
                "Formatting",
                "Measure",
                "Info",
                "Provide format string for measures",
                lambda obj, tom: obj.IsHidden is False
                and len(obj.FormatString) == 0
                and not obj.FormatStringDefinition,
                "Visible measures should have their format string property assigned.",
            ),
            (
                "Formatting",
                "Column",
                "Info",
                "Add data category for columns",
                lambda obj, tom: len(obj.DataCategory) == 0
                and any(
                    obj.Name.lower().startswith(item.lower())
                    for item in [
                        "country",
                        "city",
                        "continent",
                        "latitude",
                        "longitude",
                    ]
                ),
                "Add Data Category property for appropriate columns.",
                "https://docs.microsoft.com/power-bi/transform-model/desktop-data-categorization",
            ),
            (
                "Formatting",
                "Measure",
                "Warning",
                "Percentages should be formatted with thousands separators and 1 decimal",
                lambda obj, tom: "%" in obj.FormatString
                and obj.FormatString != "#,0.0%;-#,0.0%;#,0.0%",
                "For a better user experience, percengage measures should be formatted with a '%' sign.",
            ),
            (
                "Formatting",
                "Measure",
                "Warning",
                "Whole numbers should be formatted with thousands separators and no decimals",
                lambda obj, tom: "$" not in obj.FormatString
                and "%" not in obj.FormatString
                and obj.FormatString not in ["#,0", "#,0.0"],
                "For a better user experience, whole numbers should be formatted with commas.",
            ),
            (
                "Formatting",
                "Column",
                "Info",
                "Hide foreign keys",
                lambda obj, tom: obj.IsHidden is False
                and any(
                    r.FromColumn.Name == obj.Name
                    and r.FromCardinality == TOM.RelationshipEndCardinality.Many
                    for r in tom.used_in_relationships(object=obj)
                ),
                "Foreign keys should always be hidden as they should not be used by end users.",
            ),
            (
                "Formatting",
                "Column",
                "Info",
                "Mark primary keys",
                lambda obj, tom: any(
                    r.ToTable.Name == obj.Table.Name
                    and r.ToColumn.Name == obj.Name
                    and r.ToCardinality == TOM.RelationshipEndCardinality.One
                    for r in tom.used_in_relationships(object=obj)
                )
                and obj.IsKey is False
                and obj.Table.DataCategory != "Time",
                "Set the 'Key' property to 'True' for primary key columns within the column properties.",
            ),
            (
                "Formatting",
                "Column",
                "Info",
                "Month (as a string) must be sorted",
                lambda obj, tom: (re.search(r"month", obj.Name, flags=re.IGNORECASE))
                and not (re.search(r"months", obj.Name, flags=re.IGNORECASE))
                and (obj.DataType == TOM.DataType.String)
                and len(str(obj.SortByColumn)) == 0,
                "This rule highlights month columns which are strings and are not sorted. If left unsorted, they will sort alphabetically (i.e. April, August...). Make sure to sort such columns so that they sort properly (January, February, March...).",
            ),
            (
                "Formatting",
                "Relationship",
                "Warning",
                "Relationship columns should be of integer data type",
                lambda obj, tom: obj.FromColumn.DataType != TOM.DataType.Int64
                or obj.ToColumn.DataType != TOM.DataType.Int64,
                "It is a best practice for relationship columns to be of integer data type. This applies not only to data warehousing but data modeling as well.",
            ),
            (
                "Formatting",
                "Column",
                "Warning",
                "Provide format string for 'Month' columns",
                lambda obj, tom: re.search(r"month", obj.Name, flags=re.IGNORECASE)
                and obj.DataType == TOM.DataType.DateTime
                and obj.FormatString != "MMMM yyyy",
                'Columns of type "DateTime" that have "Month" in their names should be formatted as "MMMM yyyy".',
            ),
            (
                "Formatting",
                "Column",
                "Info",
                "Format flag columns as Yes/No value strings",
                lambda obj, tom: obj.Name.lower().startswith("is")
                and obj.DataType == TOM.DataType.Int64
                and not (obj.IsHidden or obj.Parent.IsHidden)
                or obj.Name.lower().endswith(" flag")
                and obj.DataType != TOM.DataType.String
                and not (obj.IsHidden or obj.Parent.IsHidden),
                "Flags must be properly formatted as Yes/No as this is easier to read than using 0/1 integer values.",
            ),
            (
                "Formatting",
                ["Table", "Column", "Measure", "Partition", "Hierarchy"],
                "Error",
                "Objects should not start or end with a space",
                lambda obj, tom: obj.Name[0] == " " or obj.Name[-1] == " ",
                "Objects should not start or end with a space. This usually happens by accident and is difficult to find.",
            ),
            (
                "Formatting",
                ["Table", "Column", "Measure", "Partition", "Hierarchy"],
                "Info",
                "First letter of objects must be capitalized",
                lambda obj, tom: obj.Name[0] != obj.Name[0].upper(),
                "The first letter of object names should be capitalized to maintain professional quality.",
            ),
            (
                "Naming Conventions",
                ["Table", "Column", "Measure", "Partition", "Hierarchy"],
                "Warning",
                "Object names must not contain special characters",
                lambda obj, tom: re.search(r"[\t\r\n]", obj.Name),
                "Object names should not include tabs, line breaks, etc.",
            ),
        ],
        columns=[
            "Category",
            "Scope",
            "Severity",
            "Rule Name",
            "Expression",
            "Description",
            "URL",
        ],
    )

    return rules



import pandas as pd
from uuid import UUID
from typing import Optional
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
)
import sempy_labs._icons as icons


@log
def refresh_graph(
    graph_model: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Refreshes the graph model.

    This is a wrapper function for the following API: `Background Jobs - Run On Demand Refresh Graph <https://learn.microsoft.com/rest/api/fabric/graphmodel/background-jobs/run-on-demand-refresh-graph>`_.

    Parameters
    ----------
    graph_model : str | uuid.UUID
        The graph model name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the result of the refresh operation.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=graph_model, type="GraphModel", workspace=workspace_id
    )

    print(
        f"{icons.in_progress} The refresh graph job for the '{item_name}' graph model within the '{workspace_name}' workspace has been initiated."
    )

    df = _base_api(
        request=f"/v1/workspaces/{workspace_id}/GraphModels/{item_id}/jobs/instances?jobType=RefreshGraph",
        method="post",
        lro_return_df=True,
    )
    status = df["Status"].iloc[0]

    if status == "Completed":
        print(
            f"{icons.green_dot} The refresh graph job for the '{item_name}' graph model within the '{workspace_name}' workspace has succeeded."
        )
    else:
        print(status)
        print(
            f"{icons.red_dot} The refresh graph job for the '{item_name}' graph model within the '{workspace_name}' workspace has failed."
        )

    return df



from typing import Literal, List
from uuid import UUID
import sempy_labs._icons as icons
from sempy_labs.admin._basic_functions import (
    _resolve_workspace_name_and_id,
)
from sempy_labs.admin._items import (
    list_items,
)
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
)
from sempy._utils._log import log


@log
def bulk_set_labels(
    items: List[dict],
    label_id: UUID,
    assignment_method: Literal["Standard", "Priviledged"] = "Standard",
):
    """
    Sets sensitivity labels on Fabric items.

    Note: Please use the sempy_labs.graph.resolve_sensitivity_label_id function to retrieve label IDs.

    This is a wrapper function for the following API: `Labels - Bulk Set Labels <https://learn.microsoft.com/rest/api/fabric/admin/labels/bulk-set-labels>`_.

    Parameters
    ----------
    items : List[dict]
        A list of dictionaries containing the item details.

        Example 1:
        items = [
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542a",
            "type": "Dashboard"
            },
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542c",
            "type": "Report"
            },
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542e",
            "type": "SemanticModel"
            },
        ]

        Example 2:
        items = [
            {
            "id": "Dashboard 1",
            "type": "Dashboard",
            "workspace": "Sales Workspace"
            },
            {
            "id": "Sales Report",
            "type": "Report",
            "workspace": "Sales Workspace"
            },
            {
            "id": "KPI Model",
            "type": "SemanticModel",
            "workspace": "Workspace 2"
            },
        ]

    label_id : uuid.UUID
        The label ID, which must be in the user's label policy.
    assignment_method : typing.Literal["Standard", "Priviledged"], default="Standard"
        Specifies whether the assigned label was set by an automated process or manually. Additional tenant setting property types may be added over time.
    """

    if assignment_method not in ["Standard", "Priviledged"]:
        raise ValueError("assignment_method must be either 'Standard' or 'Priviledged'")

    payload = {"items": []}
    df = list_items()

    for i in items:
        item = i.get("item")
        type = i.get("type")
        workspace = i.get("workspace")
        if _is_valid_uuid(item):
            payload["items"].append(
                {
                    "id": item,
                    "type": type,
                }
            )
        else:
            workspace_id = _resolve_workspace_name_and_id(workspace)[1]
            df_filtered = df[
                (df["Item Name"] == item)
                & (df["Type"] == type)
                & (df["Workspace Id"] == workspace_id)
            ]
            if df_filtered.empty:
                raise ValueError(
                    f"The item '{item}' of type '{type}' does not exist in workspace '{workspace}'."
                )
            else:
                payload["items"].append(
                    {
                        "id": df_filtered["Item Id"].iloc[0],
                        "type": type,
                    }
                )

    payload["labelId"] = label_id
    payload["assignmentMethod"] = assignment_method

    _base_api(request="/v1/admin/items/bulkSetLabels", method="post", payload=payload)

    print(
        f"{icons.green_dot} Labels have been successfully set on the specified items."
    )


@log
def bulk_remove_labels(
    items: List[dict],
):
    """
    Removes sensitivity labels from Fabric items.

    This is a wrapper function for the following API: `Labels - Bulk Remove Labels <https://learn.microsoft.com/rest/api/fabric/admin/labels/bulk-remove-labels>`_.

    Parameters
    ----------
    items : List[dict]
        A list of dictionaries containing the item details.

        Example 1:
        items = [
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542a",
            "type": "Dashboard"
            },
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542c",
            "type": "Report"
            },
            {
            "id": "fe472f5e-636e-4c10-a1c6-7e9edc0b542e",
            "type": "SemanticModel"
            },
        ]

        Example 2:
        items = [
            {
            "id": "Dashboard 1",
            "type": "Dashboard",
            "workspace": "Sales Workspace"
            },
            {
            "id": "Sales Report",
            "type": "Report",
            "workspace": "Sales Workspace"
            },
            {
            "id": "KPI Model",
            "type": "SemanticModel",
            "workspace": "Workspace 2"
            },
        ]
    """

    payload = {"items": []}
    df = list_items()

    for i in items:
        item = i.get("item")
        type = i.get("type")
        workspace = i.get("workspace")
        if _is_valid_uuid(item):
            payload["items"].append(
                {
                    "id": item,
                    "type": type,
                }
            )
        else:
            workspace_id = _resolve_workspace_name_and_id(workspace)[1]
            df_filtered = df[
                (df["Item Name"] == item)
                & (df["Type"] == type)
                & (df["Workspace Id"] == workspace_id)
            ]
            if df_filtered.empty:
                raise ValueError(
                    f"The item '{item}' of type '{type}' does not exist in workspace '{workspace}'."
                )
            else:
                payload["items"].append(
                    {
                        "id": df_filtered["Item Id"].iloc[0],
                        "type": type,
                    }
                )

    _base_api(
        request="/v1/admin/items/bulkRemoveLabels", method="post", payload=payload
    )

    print(
        f"{icons.green_dot} Labels have been successfully removed from the specified items."
    )



from ._items import (
    list_deployment_pipeline_operations,
    list_deployment_pipeline_role_assignments,
    list_deployment_pipeline_stage_items,
    list_deployment_pipeline_stages,
    list_deployment_pipelines,
    unassign_workspace_from_stage,
    assign_workspace_to_stage,
    delete_deployment_pipeline,
    deploy_stage_content,
)

__all__ = [
    "list_deployment_pipeline_operations",
    "list_deployment_pipeline_role_assignments",
    "list_deployment_pipeline_stage_items",
    "list_deployment_pipeline_stages",
    "list_deployment_pipelines",
    "unassign_workspace_from_stage",
    "assign_workspace_to_stage",
    "delete_deployment_pipeline",
    "deploy_stage_content",
]



from tqdm.auto import tqdm
from typing import List, Optional, Union
from sempy._utils._log import log
from uuid import UUID
from sempy_labs._helper_functions import (
    _base_api,
    resolve_lakehouse_name_and_id,
    resolve_workspace_name_and_id,
    _create_spark_session,
    _pure_python_notebook,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_workspace_id,
)
import sempy_labs._icons as icons
import re
import pandas as pd


@log
def list_lakehouses(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the lakehouses within a workspace.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the lakehouses within a workspace.
    """

    columns = {
        "Lakehouse Name": "string",
        "Lakehouse ID": "string",
        "Description": "string",
        "OneLake Tables Path": "string",
        "OneLake Files Path": "string",
        "SQL Endpoint Connection String": "string",
        "SQL Endpoint ID": "string",
        "SQL Endpoint Provisioning Status": "string",
        "Schema Enabled": "bool",
        "Default Schema": "string",
        "Sensitivity Label Id": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})
            sqlEPProp = prop.get("sqlEndpointProperties", {})
            default_schema = prop.get("defaultSchema", None)

            rows.append(
                {
                    "Lakehouse Name": v.get("displayName"),
                    "Lakehouse ID": v.get("id"),
                    "Description": v.get("description"),
                    "OneLake Tables Path": prop.get("oneLakeTablesPath"),
                    "OneLake Files Path": prop.get("oneLakeFilesPath"),
                    "SQL Endpoint Connection String": sqlEPProp.get("connectionString"),
                    "SQL Endpoint ID": sqlEPProp.get("id"),
                    "SQL Endpoint Provisioning Status": sqlEPProp.get(
                        "provisioningStatus"
                    ),
                    "Schema Enabled": True if default_schema else False,
                    "Default Schema": default_schema,
                    "Sensitivity Label Id": v.get("sensitivityLabel", {}).get(
                        "sensitivityLabelId"
                    ),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def lakehouse_attached() -> bool:
    """
    Identifies if a lakehouse is attached to the notebook.

    Returns
    -------
    bool
        Returns True if a lakehouse is attached to the notebook.
    """

    from sempy_labs._helper_functions import _get_fabric_context_setting

    lake_id = _get_fabric_context_setting(name="trident.lakehouse.id")

    if len(lake_id) > 0:
        return True
    else:
        return False


@log
def _optimize_table(path):

    if _pure_python_notebook():
        from deltalake import DeltaTable

        DeltaTable(path).optimize.compact()
    else:
        from delta import DeltaTable

        spark = _create_spark_session()
        DeltaTable.forPath(spark, path).optimize().executeCompaction()


@log
def _vacuum_table(path, retain_n_hours):

    if _pure_python_notebook():
        from deltalake import DeltaTable

        DeltaTable(path).vacuum(retention_hours=retain_n_hours)
    else:
        from delta import DeltaTable

        spark = _create_spark_session()
        spark.conf.set("spark.databricks.delta.vacuum.parallelDelete.enabled", "true")
        DeltaTable.forPath(spark, path).vacuum(retain_n_hours)


def _collect_tables(tables, lakehouse, workspace) -> pd.DataFrame:

    from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables

    df = get_lakehouse_tables(
        lakehouse=lakehouse, workspace=workspace, exclude_shortcuts=True
    )
    df_delta = df[df["Format"] == "delta"]

    if isinstance(tables, str):
        tables = [tables]

    if tables:
        filters = []

        for t in tables:
            if "." in t:
                schema, table = t.split(".", 1)
                filters.append(
                    (df_delta["Schema Name"] == schema)
                    & (df_delta["Table Name"] == table)
                )
            else:
                filters.append(df_delta["Table Name"] == t)

        # Combine all filters with OR
        combined_filter = filters[0]
        for f in filters[1:]:
            combined_filter |= f

        df_tables = df_delta[combined_filter]
    else:
        df_tables = df_delta

    df_tables.reset_index(drop=True, inplace=True)

    return df_tables


@log
def optimize_lakehouse_tables(
    tables: Optional[Union[str, List[str]]] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Runs the `OPTIMIZE <https://docs.delta.io/latest/optimizations-oss.html>`_ function over the specified lakehouse tables.

    Parameters
    ----------
    tables : str | List[str], default=None
        The table(s) to optimize. If the tables have a schema, use the 'schema.table' format.
        Defaults to None which resolves to optimizing all tables within the lakehouse.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    df_tables = _collect_tables(tables=tables, lakehouse=lakehouse, workspace=workspace)

    total = len(df_tables)
    for idx, r in (bar := tqdm(df_tables.iterrows(), total=total, bar_format="{desc}")):
        table_name = r["Table Name"]
        schema_name = r["Schema Name"]
        full_table_name = f"{schema_name}.{table_name}"
        path = r["Location"]
        bar.set_description(
            f"Optimizing the '{full_table_name if schema_name else table_name}' table ({idx + 1}/{total})..."
        )
        _optimize_table(path=path)


@log
def vacuum_lakehouse_tables(
    tables: Optional[Union[str, List[str]]] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    retain_n_hours: Optional[int] = None,
):
    """
    Runs the `VACUUM <https://docs.delta.io/latest/delta-utility.html#remove-files-no-longer-referenced-by-a-delta-table>`_ function over the specified lakehouse tables.

    Parameters
    ----------
    tables : str | List[str], default=None
        The table(s) to vacuum. If the tables have a schema, use the 'schema.table' format.
        Defaults to None which resolves to vacuuming all tables in the lakehouse.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    retain_n_hours : int, default=None
        The number of hours to retain historical versions of Delta table files.
        Files older than this retention period will be deleted during the vacuum operation.
        If not specified, the default retention period configured for the Delta table will be used.
        The default retention period is 168 hours (7 days) unless manually configured via table properties.
    """

    df_tables = _collect_tables(tables=tables, lakehouse=lakehouse, workspace=workspace)

    total = len(df_tables)
    for idx, r in (bar := tqdm(df_tables.iterrows(), total=total, bar_format="{desc}")):
        table_name = r["Table Name"]
        schema_name = r["Schema Name"]
        full_table_name = f"{schema_name}.{table_name}"
        path = r["Location"]
        bar.set_description(
            f"Vacuuming the '{full_table_name if schema_name else table_name}' table ({idx + 1}/{total})..."
        )
        _vacuum_table(path=path, retain_n_hours=retain_n_hours)


@log
def run_table_maintenance(
    table_name: str,
    optimize: bool = False,
    v_order: bool = False,
    vacuum: bool = False,
    retention_period: Optional[str] = None,
    schema: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Runs table maintenance operations on the specified table within the lakehouse.

    This is a wrapper function for the following API: `Background Jobs - Run On Demand Table Maintenance <https://learn.microsoft.com/rest/api/fabric/lakehouse/background-jobs/run-on-demand-table-maintenance>`_.

    Parameters
    ----------
    table_name : str
        Name of the delta table on which to run maintenance operations.
    optimize : bool, default=False
        If True, the `OPTIMIZE <https://docs.delta.io/latest/optimizations-oss.html>`_ function will be run on the table.
    v_order : bool, default=False
        If True, v-order will be enabled for the table.
    vacuum : bool, default=False
        If True, the `VACUUM <https://docs.delta.io/latest/delta-utility.html#remove-files-no-longer-referenced-by-a-delta-table>`_ function will be run on the table.
    retention_period : str, default=None
        If specified, the retention period for the vacuum operation. Must be in the 'd:hh:mm:ss' format.
    schema : str, default=None
        The schema of the tables within the lakehouse.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the job instance details of the table maintenance operation.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    if not optimize and not vacuum:
        raise ValueError(
            f"{icons.warning} At least one of 'optimize' or 'vacuum' must be set to True."
        )
    if not vacuum and retention_period is not None:
        raise ValueError(
            f"{icons.warning} The 'retention_period' parameter can only be set if 'vacuum' is set to True."
        )
    if retention_period is not None:

        def is_valid_format(time_string):
            pattern = r"^\d+:[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$"
            return bool(re.match(pattern, time_string))

        if not is_valid_format(retention_period):
            raise ValueError(
                f"{icons.red_dot} The 'retention_period' parameter must be in the 'd:hh:mm:ss' format."
            )

    payload = {
        "executionData": {
            "tableName": table_name,
        }
    }
    if schema is not None:
        payload["executionData"]["schemaName"] = schema
    if optimize:
        payload["executionData"]["optimizeSettings"] = {}
    if v_order:
        payload["executionData"]["optimizeSettings"] = {"vOrder": True}
    if vacuum:
        payload["executionData"]["vacuumSettings"] = {}
    if vacuum and retention_period is not None:
        payload["executionData"]["vacuumSettings"]["retentionPeriod"] = retention_period

    print(
        f"{icons.in_progress} The table maintenance job for the '{table_name}' table in the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has been initiated."
    )

    df = _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/jobs/instances?jobType=TableMaintenance",
        method="post",
        payload=payload,
        status_codes=[200, 202],
        client="fabric_sp",
        lro_return_df=True,
    )

    status = df["Status"].iloc[0]

    if status == "Completed":
        print(
            f"{icons.green_dot} The table maintenance job for the '{table_name}' table in the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has succeeded."
        )
    else:
        print(status)
        print(
            f"{icons.red_dot} The table maintenance job for the '{table_name}' table in the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has failed."
        )

    return df



import pandas as pd
from typing import Literal, Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs.connection as conn


@log
def delete_connection(connection: str | UUID):
    """
    Delete a connection.

    This is a wrapper function for the following API: `Connections - Delete Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/delete-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.
    """

    conn.delete_connection(connection)


@log
def delete_connection_role_assignment(connection: str | UUID, role_assignment_id: UUID):
    """
    Delete the specified role assignment for the connection.

    This is a wrapper function for the following API: `Connections - Delete Connection Role Assignment <https://learn.microsoft.com/rest/api/fabric/core/connections/delete-connection-role-assignment>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.
    role_assignment_id : uuid.UUID
        The role assignment ID.
    """

    conn.delete_connection_role_assignment(connection, role_assignment_id)


@log
def resolve_connection_id(connection: str | UUID) -> UUID:

    return conn.resolve_connection_id(connection)


@log
def list_connection_role_assignments(connection: str | UUID) -> pd.DataFrame:
    """
    Returns a list of connection role assignments.

    This is a wrapper function for the following API: `Connections - List Connection Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/connections/list-connection-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of connection role assignments.
    """

    return conn.list_connection_role_assignments(connection)


@log
def list_connections() -> pd.DataFrame:
    """
    Lists all available connections.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all available connections.
    """

    return conn.list_connections()


@log
def list_item_connections(
    item: Optional[str | UUID] = None,
    type: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Shows the list of connections that the specified item is connected to.

    This is a wrapper function for the following API: `Items - List Item Connections <https://learn.microsoft.com/rest/api/fabric/core/items/list-item-connections>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The item name or ID.
    type : str
        The `item type <https://learn.microsoft.com/rest/api/fabric/core/items/update-item?tabs=HTTP#itemtype>`_.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the list of connections that the specified item is connected to.
    """

    return conn.list_item_connections(
        item=item, type=type, workspace=workspace, **kwargs
    )


@log
def list_supported_connection_types(
    gateway: Optional[str | UUID] = None, show_all_creation_methods: bool = False
) -> pd.DataFrame:

    return conn.list_supported_connection_types(
        gateway=gateway, show_all_creation_methods=show_all_creation_methods
    )


@log
def create_cloud_connection(
    name: str,
    server_name: str,
    database_name: str,
    user_name: str,
    password: str,
    privacy_level: str,
    connection_encryption: Literal["Encrypted", "Any", "NotEncrypted"] = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates a shared cloud connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    user_name : str
        The username.
    password : str
        The password.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : typing.Literal["Encrypted", "Any", "NotEncrypted"], default="NotEncrypted"
        The connection encryption type.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    conn.create_cloud_connection(
        name=name,
        server_name=server_name,
        database_name=database_name,
        user_name=user_name,
        password=password,
        privacy_level=privacy_level,
        connection_encryption=connection_encryption,
        skip_test_connection=skip_test_connection,
    )


@log
def create_on_prem_connection(
    name: str,
    gateway: str | UUID,
    server_name: str,
    database_name: str,
    credentials: str,
    privacy_level: str,
    connection_encryption: Literal["Encrypted", "Any", "NotEncrypted"] = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates an on-premises connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    gateway : str | uuid.UUID
        The name or Id of the gateway.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    credentials : str
        The encrypted credentials obtained from the gateway.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : typing.Literal["Encrypted", "Any", "NotEncrypted"], default="NotEncrypted"
        The connection encryption type.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    conn.create_on_prem_connection(
        name=name,
        gateway=gateway,
        server_name=server_name,
        database_name=database_name,
        credentials=credentials,
        privacy_level=privacy_level,
        connection_encryption=connection_encryption,
        skip_test_connection=skip_test_connection,
    )


@log
def create_vnet_connection(
    name: str,
    gateway: str | UUID,
    server_name: str,
    database_name: str,
    user_name: str,
    password: str,
    privacy_level: str,
    connection_encryption: Literal["Encrypted", "Any", "NotEncrypted"] = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates a virtual network gateway connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    gateway : str | uuid.UUID
        The name or Id of the gateway.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    user_name : str
        The username.
    password : str
        The password.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : typing.Literal["Encrypted", "Any", "NotEncrypted"], default="NotEncrypted"
        The connection encryption type.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    conn.create_vnet_connection(
        name=name,
        gateway=gateway,
        server_name=server_name,
        database_name=database_name,
        user_name=user_name,
        password=password,
        privacy_level=privacy_level,
        connection_encryption=connection_encryption,
        skip_test_connection=skip_test_connection,
    )



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    resolve_item_name_and_id,
    resolve_lakehouse_name_and_id,
    resolve_workspace_id,
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    resolve_workspace_name,
)
from sempy._utils._log import log
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from sempy.fabric.exceptions import FabricHTTPException


@log
def create_shortcut_onelake(
    table_name: str,
    source_workspace: str | UUID,
    destination_lakehouse: Optional[str | UUID] = None,
    destination_workspace: Optional[str | UUID] = None,
    shortcut_name: Optional[str] = None,
    source_item: str | UUID = None,
    source_item_type: str = "Lakehouse",
    source_path: str = "Tables",
    destination_path: str = "Tables",
    shortcut_conflict_policy: Optional[str] = None,
    **kwargs,
):
    """
    Creates a `shortcut <https://learn.microsoft.com/fabric/onelake/onelake-shortcuts>`_ to a delta table in OneLake.

    This is a wrapper function for the following API: `OneLake Shortcuts - Create Shortcut <https://learn.microsoft.com/rest/api/fabric/core/onelake-shortcuts/create-shortcut>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    table_name : str
        The table name for which a shortcut will be created.
    source_workspace : str | uuid.UUID
        The name or ID of the Fabric workspace in which the source data store exists.
    destination_lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse in which the shortcut will be created.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    destination_workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the shortcut will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    shortcut_name : str, default=None
        The name of the shortcut 'table' to be created. This defaults to the 'table_name' parameter value.
    source_item : str | uuid.UUID, default=None
        The source Fabric data store item in which the table resides. Can be either the Name or ID of the item.
    source_item_type: str, default="Lakehouse"
        The source Fabric data store item type. Options are 'Lakehouse', 'Warehouse', 'MirroredDatabase', 'SQLDatabase', and 'KQLDatabase'.
    source_path : str, default="Tables"
        A string representing the full path to the table/file in the source lakehouse, including either "Files" or "Tables". Examples: Tables/FolderName/SubFolderName; Files/FolderName/SubFolderName.
    destination_path: str, default="Tables"
        A string representing the full path where the shortcut is created, including either "Files" or "Tables". Examples: Tables/FolderName/SubFolderName; Files/FolderName/SubFolderName.
    shortcut_conflict_policy : str, default=None
        When provided, it defines the action to take when a shortcut with the same name and path already exists. The default action is 'Abort'. Additional ShortcutConflictPolicy types may be added over time.
    """

    if source_item is None:
        if "source_lakehouse" in kwargs:
            source_item = kwargs.get("source_lakehouse")
        else:
            raise ValueError(
                f"{icons.red_dot} The 'source_item' parameter must be provided."
            )

    if not (source_path.startswith("Files") or source_path.startswith("Tables")):
        raise ValueError(
            f"{icons.red_dot} The 'source_path' parameter must be either 'Files' or 'Tables'."
        )
    if not (
        destination_path.startswith("Files") or destination_path.startswith("Tables")
    ):
        raise ValueError(
            f"{icons.red_dot} The 'destination_path' parameter must be either 'Files' or 'Tables'."
        )
    if not (
        source_item_type
        in ["Lakehouse", "Warehouse", "MirroredDatabase", "SQLDatabase", "KQLDatabase"]
    ):
        raise ValueError(
            f"{icons.red_dot} The 'source_item_type' parameter must be 'Lakehouse', 'Warehouse', 'MirroredDatabase', 'SQLDatabase', or 'KQLDatabase'"
        )

    (source_workspace_name, source_workspace_id) = resolve_workspace_name_and_id(
        source_workspace
    )

    (source_item_name, source_item_id) = resolve_item_name_and_id(
        item=source_item, type=source_item_type, workspace=source_workspace_id
    )

    (destination_workspace_name, destination_workspace_id) = (
        resolve_workspace_name_and_id(destination_workspace)
    )
    (destination_lakehouse_name, destination_lakehouse_id) = (
        resolve_lakehouse_name_and_id(
            lakehouse=destination_lakehouse, workspace=destination_workspace_id
        )
    )

    if shortcut_name is None:
        shortcut_name = table_name

    source_full_path = f"{source_path}/{table_name}"

    actual_shortcut_name = shortcut_name.replace(" ", "")

    payload = {
        "path": destination_path,
        "name": actual_shortcut_name,
        "target": {
            "oneLake": {
                "workspaceId": source_workspace_id,
                "itemId": source_item_id,
                "path": source_full_path,
            }
        },
    }

    # Check if the shortcut already exists
    try:
        response = _base_api(
            request=f"/v1/workspaces/{destination_workspace_id}/items/{destination_lakehouse_id}/shortcuts/{destination_path}/{actual_shortcut_name}",
            client="fabric_sp",
        )
        response_json = response.json()
        del response_json["target"]["type"]
        if response_json.get("target") == payload.get("target"):
            print(
                f"{icons.info} The '{actual_shortcut_name}' shortcut already exists in the '{destination_lakehouse_name}' lakehouse within the '{destination_workspace_name}' workspace."
            )
            return
        else:
            raise ValueError(
                f"{icons.red_dot} The '{actual_shortcut_name}' shortcut already exists in the '{destination_lakehouse_name} lakehouse within the '{destination_workspace_name}' workspace but has a different source."
            )
    except FabricHTTPException:
        pass

    url = f"/v1/workspaces/{destination_workspace_id}/items/{destination_lakehouse_id}/shortcuts"

    if shortcut_conflict_policy:
        if shortcut_conflict_policy not in ["Abort", "GenerateUniqueName"]:
            raise ValueError(
                f"{icons.red_dot} The 'shortcut_conflict_policy' parameter must be either 'Abort' or 'GenerateUniqueName'."
            )
        url += f"?shortcutConflictPolicy={shortcut_conflict_policy}"

    _base_api(
        request=url,
        payload=payload,
        status_codes=201,
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The shortcut '{shortcut_name}' was created in the '{destination_lakehouse_name}' lakehouse within the '{destination_workspace_name}' workspace. It is based on the '{table_name}' table in the '{source_item_name}' {source_item_type} within the '{source_workspace_name}' workspace."
    )


def create_shortcut(
    shortcut_name: str,
    location: str,
    subpath: str,
    source: str,
    connection_id: str,
    lakehouse: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """
    Creates a `shortcut <https://learn.microsoft.com/fabric/onelake/onelake-shortcuts>`_ to an ADLS Gen2 or Amazon S3 source.

    Parameters
    ----------
    shortcut_name : str
    location : str
    subpath : str
    source : str
    connection_id: str
    lakehouse : str
        The Fabric lakehouse in which the shortcut will be created.
    workspace : str, default=None
        The name of the Fabric workspace in which the shortcut will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    source_titles = {"adlsGen2": "ADLS Gen2", "amazonS3": "Amazon S3"}

    sourceValues = list(source_titles.keys())

    if source not in sourceValues:
        raise ValueError(
            f"{icons.red_dot} The 'source' parameter must be one of these values: {sourceValues}."
        )

    sourceTitle = source_titles[source]

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    shortcutActualName = shortcut_name.replace(" ", "")

    payload = {
        "path": "Tables",
        "name": shortcutActualName,
        "target": {
            source: {
                "location": location,
                "subpath": subpath,
                "connectionId": connection_id,
            }
        },
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{lakehouse_id}/shortcuts",
        method="post",
        payload=payload,
        status_codes=201,
        client="fabric_sp",
    )
    print(
        f"{icons.green_dot} The shortcut '{shortcutActualName}' was created in the '{lakehouse_name}' lakehouse within"
        f" the '{workspace_name}' workspace. It is based on the '{subpath}' table in '{sourceTitle}'."
    )


@log
def delete_shortcut(
    shortcut_name: str,
    shortcut_path: str = "Tables",
    lakehouse: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Deletes a shortcut.

    This is a wrapper function for the following API: `OneLake Shortcuts - Delete Shortcut <https://learn.microsoft.com/rest/api/fabric/core/onelake-shortcuts/delete-shortcut>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    shortcut_name : str
        The name of the shortcut.
    shortcut_path : str = "Tables"
        The path of the shortcut to be deleted. Must start with either "Files" or "Tables". Examples: Tables/FolderName/SubFolderName; Files/FolderName/SubFolderName.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name in which the shortcut resides.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | UUID, default=None
        The name or ID of the Fabric workspace in which lakehouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{lakehouse_id}/shortcuts/{shortcut_path}/{shortcut_name}",
        method="delete",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{shortcut_name}' shortcut in the '{lakehouse}' within the '{workspace_name}' workspace has been deleted."
    )


@log
def reset_shortcut_cache(workspace: Optional[str | UUID] = None):
    """
    Deletes any cached files that were stored while reading from shortcuts.

    This is a wrapper function for the following API: `OneLake Shortcuts - Reset Shortcut Cache <https://learn.microsoft.com/rest/api/fabric/core/onelake-shortcuts/reset-shortcut-cache>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/onelake/resetShortcutCache",
        method="post",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The shortcut cache has been reset for the '{workspace_name}' workspace."
    )


@log
def list_shortcuts(
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    path: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows all shortcuts which exist in a Fabric lakehouse and their properties.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which lakehouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    path: str, default=None
        The path within lakehouse where to look for shortcuts. If provided, must start with either "Files" or "Tables". Examples: Tables/FolderName/SubFolderName; Files/FolderName/SubFolderName.
        Defaults to None which will retun all shortcuts on the given lakehouse

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all the shortcuts which exist in the specified lakehouse.
    """

    workspace_id = resolve_workspace_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    columns = {
        "Shortcut Name": "string",
        "Shortcut Path": "string",
        "Source Type": "string",
        "Source Workspace Id": "string",
        "Source Workspace Name": "string",
        "Source Item Id": "string",
        "Source Item Name": "string",
        "Source Item Type": "string",
        "OneLake Path": "string",
        "Connection Id": "string",
        "Location": "string",
        "Bucket": "string",
        "SubPath": "string",
        "Source Properties Raw": "string",
    }
    df = _create_dataframe(columns=columns)

    # To improve performance create a dataframe to cache all items for a given workspace
    itm_clms = {
        "Id": "string",
        "Display Name": "string",
        "Description": "string",
        "Type": "string",
        "Workspace Id": "string",
    }
    source_items_df = _create_dataframe(columns=itm_clms)

    url = f"/v1/workspaces/{workspace_id}/items/{lakehouse_id}/shortcuts"

    if path is not None:
        url += f"?parentPath={path}"

    responses = _base_api(
        request=url,
        uses_pagination=True,
        client="fabric_sp",
    )

    sources = {
        "AdlsGen2": "adlsGen2",
        "AmazonS3": "amazonS3",
        "Dataverse": "dataverse",
        "ExternalDataShare": "externalDataShare",
        "GoogleCloudStorage": "googleCloudStorage",
        "OneLake": "oneLake",
        "S3Compatible": "s3Compatible",
    }

    rows = []
    for r in responses:
        for i in r.get("value", []):
            tgt = i.get("target", {})
            tgt_type = tgt.get("type")
            connection_id = tgt.get(sources.get(tgt_type), {}).get("connectionId")
            location = tgt.get(sources.get(tgt_type), {}).get("location")
            sub_path = tgt.get(sources.get(tgt_type), {}).get("subpath")
            source_workspace_id = tgt.get(sources.get(tgt_type), {}).get("workspaceId")
            source_item_id = tgt.get(sources.get(tgt_type), {}).get("itemId")
            bucket = tgt.get(sources.get(tgt_type), {}).get("bucket")
            source_workspace_name = (
                resolve_workspace_name(
                    workspace_id=source_workspace_id, throw_error=False
                )
                if source_workspace_id is not None
                else None
            )
            # Cache and use it to getitem type and name
            source_item_type = None
            source_item_name = None
            try:
                dfI = source_items_df[
                    source_items_df["Workspace Id"] == source_workspace_id
                ]
                if dfI.empty:
                    dfI = fabric.list_items(workspace=source_workspace_id)
                    source_items_df = pd.concat(
                        [source_items_df, dfI], ignore_index=True
                    )

                dfI_filt = dfI[dfI["Id"] == source_item_id]
                if not dfI_filt.empty:
                    source_item_type = dfI_filt["Type"].iloc[0]
                    source_item_name = dfI_filt["Display Name"].iloc[0]
            except Exception:
                pass

            rows.append(
                {
                    "Shortcut Name": i.get("name"),
                    "Shortcut Path": i.get("path"),
                    "Source Type": tgt_type,
                    "Source Workspace Id": source_workspace_id,
                    "Source Workspace Name": source_workspace_name,
                    "Source Item Id": source_item_id,
                    "Source Item Name": source_item_name,
                    "Source Item Type": source_item_type,
                    "OneLake Path": tgt.get(sources.get("oneLake"), {}).get("path"),
                    "Connection Id": connection_id,
                    "Location": location,
                    "Bucket": bucket,
                    "SubPath": sub_path,
                    "Source Properties Raw": str(tgt),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



import pandas as pd
from typing import Optional, List
from sempy._utils._log import log
from os import PathLike
from uuid import UUID
import sempy_labs.notebook as nb


def _get_notebook_definition_base(
    notebook_name: str,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
) -> pd.DataFrame:

    return nb.get_notebook_definition_base(
        notebook_name=notebook_name, workspace=workspace, format=format
    )


def get_notebook_type(
    notebook_name: str, workspace: Optional[str | UUID] = None
) -> str:

    return nb.get_notebook_type(notebook_name=notebook_name, workspace=workspace)


@log
def get_notebook_definition(
    notebook_name: str,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
    format: Optional[str] = None,
) -> str:
    """
    Obtains the notebook definition.

    This is a wrapper function for the following API: `Items - Get Notebook Definition <https://learn.microsoft.com/rest/api/fabric/notebook/items/get-notebook-definition>`_.

    Parameters
    ----------
    notebook_name : str
        The name of the notebook.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        If True, decodes the notebook definition file into .ipynb format.
        If False, obtains the notebook definition file in base64 format.
    format : str, default=None
        The only supported value is ipynb
        If provided the format will be in standard .ipynb otherwise the format will be in source code format which is GIT friendly ipynb

    Returns
    -------
    str
        The notebook definition.
    """

    return nb.get_notebook_definition(
        notebook_name=notebook_name,
        workspace=workspace,
        decode=decode,
        format=format,
    )


@log
def import_notebook_from_web(
    notebook_name: str,
    url: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    overwrite: bool = False,
    folder: Optional[str | PathLike] = None,
):
    """
    Creates a new notebook within a workspace based on a Jupyter notebook hosted in the web.

    Note: When specifying a notebook from GitHub, please use the raw file path. Note that if the non-raw file path is specified, the url will be
    converted to the raw URL as the raw URL is needed to obtain the notebook content.

    Parameters
    ----------
    notebook_name : str
        The name of the notebook to be created.
    url : str
        The url of the Jupyter Notebook (.ipynb)
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    overwrite : bool, default=False
        If set to True, overwrites the existing notebook in the workspace if it exists.
    folder : str | os.PathLike, default=None
        The folder within the workspace where the notebook will be created.
        Defaults to None which places the notebook in the root of the workspace.
    """

    nb.import_notebook_from_web(
        notebook_name=notebook_name,
        url=url,
        description=description,
        workspace=workspace,
        overwrite=overwrite,
        folder=folder,
    )


@log
def create_notebook(
    name: str,
    notebook_content: str,
    type: str = "py",
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
    folder: Optional[str | PathLike] = None,
):
    """
    Creates a new notebook with a definition within a workspace.

    Parameters
    ----------
    name : str
        The name of the notebook to be created.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    type : str, default="py"
        The notebook type.
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    format : str, default=None
        If 'ipynb' is provided than notebook_content should be standard ipynb format
        otherwise notebook_content should be GIT friendly format
    folder : str | os.PathLike, default=None
        The folder within the workspace where the notebook will be created.
        Defaults to None which places the notebook in the root of the workspace.
    """

    nb.create_notebook(
        name=name,
        notebook_content=notebook_content,
        type=type,
        description=description,
        workspace=workspace,
        format=format,
        folder=folder,
    )


@log
def update_notebook_definition(
    name: str,
    notebook_content: str,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
):
    """
    Updates an existing notebook with a new definition.

    Parameters
    ----------
    name : str
        The name of the notebook to be updated.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    format : str, default=None
        If 'ipynb' is provided than notebook_content should be standard ipynb format
        otherwise notebook_content should be GIT friendly format
    """

    nb.update_notebook_definition(
        name=name,
        notebook_content=notebook_content,
        workspace=workspace,
        format=format,
    )


@log
def list_notebooks(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the notebooks within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the notebooks within a workspace.
    """

    return nb.list_notebooks(workspace=workspace)


@log
def search_notebooks(
    search_string: str,
    notebook: Optional[str | UUID] = None,
    workspace: Optional[str | UUID | List[str | UUID]] = None,
) -> pd.DataFrame:
    """
    Searches notebooks within a workspace or across multiple workspaces for a given search string.

    Parameters
    ----------
    search_string : str
        The string to search for within the notebook definitions.
    notebook : str | uuid.UUID, default=None
        The name or ID of a specific notebook to search within.
        Defaults to None which searches across all notebooks in the specified workspace(s).
    workspace : str | uuid.UUID | list, default=None
        The name or ID of the workspace or a list of workspaces to search within.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
        If a list is provided, it should contain workspace names or IDs.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the notebooks that contain the search string in their definitions.
        The dataframe includes the workspace name, workspace ID, notebook name, and notebook ID.
    """

    return nb.search_notebooks(
        search_string=search_string,
        notebook=notebook,
        workspace=workspace,
    )



from uuid import UUID
import sempy_labs._icons as icons
import pandas as pd
from sempy_labs.admin._basic_functions import _resolve_workspace_name_and_id
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
)
from sempy._utils._log import log


@log
def list_external_data_shares() -> pd.DataFrame:
    """
    Lists external data shares in the tenant. This function is for admins.

    This is a wrapper function for the following API: `External Data Shares - List External Data Shares <https://learn.microsoft.com/rest/api/fabric/admin/external-data-shares/list-external-data-shares>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of external data shares in the tenant.
    """

    columns = {
        "External Data Share Id": "string",
        "Paths": "string",
        "Creater Principal Id": "string",
        "Creater Principal Name": "string",
        "Creater Principal Type": "string",
        "Creater Principal UPN": "string",
        "Recipient UPN": "string",
        "Status": "string",
        "Expiration Time UTC": "datetime",
        "Workspace Id": "string",
        "Item Id": "string",
        "Invitation URL": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="/v1/admin/items/externalDataShares")

    rows = []
    for i in response.json().get("value", []):
        cp = i.get("creatorPrincipal", {})
        rows.append(
            {
                "External Data Share Id": i.get("id"),
                "Paths": [i.get("paths", [])],
                "Creater Principal Id": cp.get("id"),
                "Creater Principal Name": cp.get("displayName"),
                "Creater Principal Type": cp.get("type"),
                "Creater Principal UPN": cp.get("userDetails", {}).get(
                    "userPrincipalName"
                ),
                "Recipient UPN": i.get("recipient", {}).get("userPrincipalName"),
                "Status": i.get("status"),
                "Expiration Time UTC": i.get("expirationTimeUtc"),
                "Workspace Id": i.get("workspaceId"),
                "Item Id": i.get("itemId"),
                "Invitation URL": i.get("invitationUrl"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def revoke_external_data_share(
    external_data_share_id: UUID, item_id: UUID, workspace: str | UUID
):
    """
    Revokes the specified external data share. Note: This action cannot be undone.

    This is a wrapper function for the following API: `External Data Shares - Revoke External Data Share <https://learn.microsoft.com/rest/api/fabric/admin/external-data-shares/revoke-external-data-share>`_.

    Parameters
    ----------
    external_data_share_id : uuid.UUID
        The external data share ID.
    item_id : uuid.UUID, default=None
        The Item ID
    workspace : str | uuid.UUID
        The Fabric workspace name or id.
    """
    (workspace, workspace_id) = _resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1/admin/workspaces/{workspace_id}/items/{item_id}/externalDataShares/{external_data_share_id}/revoke",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{external_data_share_id}' external data share for the '{item_id}' item within the '{workspace}' workspace has been revoked."
    )
